/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: core_cache.c,v 1.2 2002/01/13 21:33:36 zarjazz Exp $
	----
	Description:

	A Tree based IP / Hostname Cache implementation.
*/

#include <sys_include.h>
#include <sys_common.h>
#include <sys_string.h>

#include <core_threads.h>
#include <core_cache.h>

/* ================ */

/*
 * my own fixed-time ip / subnet connection count lookup, with dns host cache
 * the algorithm seemed liked a good idea at the time but god knows if its efficient
 * (memory fragmentation could be a bitch! :)
 */

static u_int32_t _branches;
static ip_tree_t *_root;

rwlock_t rwlock_ip;

/* ================ */

static void __ip_cleanup (ip_tree_t *tree, int bit)
{
	if (!tree) return;
	if (bit--)
	{
		__ip_cleanup(tree->right,bit);
		__ip_cleanup(tree->left ,bit);
	}
	/* cleanup the dns cache */
	else
	{
		sys_free(tree->left);
		sys_free(tree->right);
	}
	sys_free(tree);
}

static void __ip_del (ip_tree_t *tree, u_int32_t ip, int bit)
{
	if (!tree) return;

	tree->count--;
	if (bit)
	{
		ip_tree_t **branch = (ip & bit) ? &tree->right : &tree->left;

		__ip_del(*branch, ip, bit - 1);
		*branch = NULL;
	}

	if (!tree->count)
	{
		/* cleanup the dns cache */
		if (!bit) { sys_free(tree->left); sys_free(tree->right); }
		/* free the tree unless root branch */
		if (tree != _root) { sys_free(tree); _branches--; }
	}
}

static ip_tree_t * __ip_add (ip_tree_t *tree, u_int32_t ip, u_int32_t tv_sec, int bit)
{
	assert(tree);

	tree->tv_sec = tv_sec;
	tree->count++;
	if (bit)
	{
		ip_tree_t **branch = (ip & bit) ? &tree->right : &tree->left;

		if (*branch == NULL) { *branch = sys_malloc(sizeof(ip_tree_t)); _branches++; }
		return __ip_add(*branch, ip, tv_sec, bit - 1);
	}
	return tree;
}

static ip_tree_t * __ip_match (ip_tree_t *tree, u_int32_t ip, int bit, int lvl)
{
	ip_tree_t *ret = NULL;

	if (bit && bit > lvl)
	{
		ip_tree_t *branch = (ip & bit) ? tree->right : tree->left;

		if (branch)	ret = __ip_match(branch, ip, bit - 1, lvl);
	}
	else
		ret = tree;

	return ret;
}

/*
================
ip_add
================
*/

ip_tree_t * ip_add (u_int32_t ip, u_int32_t tv_sec, bool lock)
{
	ip_tree_t *ret;
	if (lock) WRITE_LOCK(rwlock_ip);
	ret = __ip_add(_root,ip,tv_sec,31);
	if (lock) WRITE_UNLOCK(rwlock_ip);
	return ret;
}

/*
================
ip_del
================
*/

void ip_del (u_int32_t ip)
{
	WRITE_LOCK(rwlock_ip);
	__ip_del(_root,ip,31);
	WRITE_UNLOCK(rwlock_ip);
}

/*
================
ip_count
================
*/

u_int32_t ip_branches (void) { return _branches; }

/*
================
ip_find
================
*/

ip_tree_t * ip_find (u_int32_t ip)
{
	ip_tree_t *tree;
	READ_LOCK(rwlock_ip);
	tree = __ip_match(_root,ip,31,0);
	READ_UNLOCK(rwlock_ip);
	return tree;
}

/*
================
ip_count
================
*/

u_int32_t ip_count (u_int32_t ip, int subnet)
{
	int count = 0;
	ip_tree_t *tree;

	READ_LOCK(rwlock_ip);
	tree = __ip_match(_root,ip,31,32 - subnet);
	if (tree) count = tree->count;
	READ_UNLOCK(rwlock_ip);
	return count;
}

/* ================ */

/*
================
cleanup_ip_cache
================
*/

static void cleanup_ip_cache (void)
{
	sys_dprintf(2,"cleaning IP cache\n");

	/* ================ */

	WRITE_LOCK(rwlock_ip);
	__ip_cleanup(_root,31);
	WRITE_UNLOCK(rwlock_ip);

	RWLOCK_DESTROY(rwlock_ip);

	/* ================ */

	sys_dprintf(2,"IP cache cleaned\n");
}

/*
================
init_ip_cache
================
*/

void init_ip_cache (void)
{
	_branches = 1;
	_root = sys_malloc(sizeof(ip_tree_t));
	RWLOCK_INIT(rwlock_ip);

	/* register for cleanup */
	sys_cleanup_register(cleanup_ip_cache);
}
