/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: core_node.c,v 1.4 2002/01/18 20:47:40 zarjazz Exp $
*/

#include <sys_include.h>

#include <core_node.h>

/* ================ */

static unsigned MAX_NODES;
static node_t **RouteTable;
node_t *Node;

/*
================
init_nodes
================
*/

void init_nodes (node_t *node, unsigned max_nodes)
{
	Node = node;
	MAX_NODES = max_nodes;
}

/*
================
node_find_route
================
*/

node_t *node_find_route (u_int16_t id) { return RouteTable[Node->route[id]]; }

/*
================
link_add
================
*/

void link_add (node_t *node, u_int16_t id)
{
	int i;

	for (i = 0; i < node->num_neighbours; i++)
		if (node->neighbours[i] == id) return;

	node->neighbours[node->num_neighbours++] = id;
}

/*
================
link_del
================
*/

void link_del (node_t *node, u_int16_t id)
{
	int i;

	for (i = 0; i < node->num_neighbours; i++)
		if (node->neighbours[i] == id)
		{
			node->neighbours[i] = node->neighbours[--node->num_neighbours];
			return;
		}
}

/*
================
link_state_string
----------------
build the current link state of a node
================
*/

void link_state_string (char *dst, node_t *node)
{
	int i;

	for (i = 0; i < node->num_neighbours; i++)
		dst += sprintf(dst," +%hu", node->neighbours[i]);
}

/*
================
calculate_cost
================
*/

static bool calculate_cost (node_t *src, node_t *dst, node_t **list, int *cost)
{
	u_int16_t i;
	int best_cost = -1;

	/* check if 'dst' is a neighbour */
	for (i = 0; i < src->num_neighbours; i++)
		if (src->neighbours[i] == dst->id) { *cost = *cost + 1; return true; }

	/* remove hop from list */
	list[src->id] = NULL;

	/* check route via each neighbour */
	for (i = 0; i < src->num_neighbours; i++)
	{
		node_t *node;
		u_int16_t id;
		bool linked;
		int rt_cost;

		rt_cost = *cost + 1;
		id = src->neighbours[i];
		node = list[id];
		/* if we have we already visited this node, then skip */
		if (!node) continue;

		linked = calculate_cost(node, dst, list, &rt_cost);
		if (linked && (best_cost < 0 || rt_cost < best_cost)) best_cost = rt_cost;
	}

	/* replace hop */
	list[src->id] = RouteTable[src->id];

	/* can we reach the destination? */
	if (best_cost < 0) return false;

	*cost = best_cost;
	return true;
}

/*
================
calculate_shortest_path
================
*/

static u_int16_t calculate_shortest_path (node_t *src, node_t *dst, node_t **list)
{
	u_int16_t i, best_id = 0;
	int best_cost = -1;

	/* check if 'dst' is a neighbour */
	for (i = 0; i < src->num_neighbours; i++)
		if (src->neighbours[i] == dst->id) return 0;

	/* remove root node from list */
	list[src->id] = NULL;

	/* check route via each neighbour */
	for (i = 0; i < src->num_neighbours; i++)
	{
		node_t *node;
		u_int16_t id;
		bool linked;
		int rt_cost;

		rt_cost = 0;
		id = src->neighbours[i];
		node = RouteTable[id];

		linked = calculate_cost(node, dst, list, &rt_cost);
		if (linked && (best_cost < 0 || rt_cost < best_cost)) { best_id = id; best_cost = rt_cost; }
	}

	/* replace root node */
	list[src->id] = RouteTable[src->id];

	return best_id;
}

/*
================
calculate_best_routes
================
*/

void calculate_best_routes (void)
{
	node_t **list = alloca(MAX_NODES * sizeof(node_t *));
	node_t *src_node, *dst_node;
	u_int16_t x, y;

	/* create a copy of the RouteTable */
	memcpy(list, RouteTable, MAX_NODES * sizeof(node_t *));

	/* cycle through all nodes */
	for (x = 1; x < MAX_NODES; x++)
	{
		src_node = RouteTable[x];
		if (!src_node) continue;

		for (y = 1; y < MAX_NODES; y++)
		{
			if (x == y) continue;

			dst_node = RouteTable[y];
			if (!dst_node) continue;

			/* calculate best route */
			src_node->route[dst_node->id] = calculate_shortest_path(src_node, dst_node, list);
		}
	}
}
