/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: core_tree.c,v 1.3 2002/01/13 21:33:36 zarjazz Exp $
	----
	Description:

	A string comparision based Binary Tree implementation.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <core_tree.h>

/* ================ */

#ifndef WINVER
  #define stricmp strcasecmp
#endif

/*
================
tree_find
================
*/

tree_t * tree_find (tree_t *tree, char *key)
{
	int rc;

	if (!tree) return NULL;

	rc = stricmp(tree->key,key);
	if (rc == 0) return tree;
	return tree_find(rc < 0 ? tree->left : tree->right, key);
}

/*
================
tree_add
================
*/

static tree_t * __tree_add (tree_t *tree, tree_t *node)
{
	int rc = stricmp(tree->key,node->key);

	if (rc < 0)
	{
		/* left */
		if (tree->left) return __tree_add(tree->left,node);
		tree->left = node;
	}
	else if (rc > 0)
	{
		/* right */
		if (tree->right) return __tree_add(tree->right,node);
		tree->right = node;
	}
	else
		/* we already have a match, bad! */
#ifdef _DEBUG
		assert(1 == 0);
#else
		return NULL;
#endif /* _DEBUG */

	node->parent = tree;
	return node;
}

tree_t * tree_add (tree_t **root, tree_t *node)
{
	if (*root)
		return __tree_add(*root, node);
	else
		return (*root = node);
}

/*
================
tree_del
================
*/

void tree_del (tree_t **root, tree_t *node)
{
	tree_t **branch;

	if (!node->parent)
		/* delete root */
		branch = root;
	else if (node->parent->left == node)
		/* delete left branch */
		branch = &node->parent->left;
	else if (node->parent->right == node)
		/* delete right branch */
		branch = &node->parent->right;
	else
	{
		assert(1 == 0);
		return;
	}

	/* relink branch */
	if (node->left && node->right)
	{
		tree_t *relink = node->left;
		/* relink immediate predecessor to right branch */
		while (relink->right) relink = relink->right;

		if (relink != node->left)
		{
			if (relink->parent->right = relink->left)
				relink->left->parent = relink->parent;
			relink->left = node->left;
			relink->left->parent = relink;
		}

		relink->right	= node->right;
		relink->parent	= node->parent;

		relink->right->parent = relink;

		*branch = relink;
	}
	else if (node->left)
	{
		node->left->parent = node->parent;
		*branch = node->left;
	}
	else if (node->right)
	{
		node->right->parent = node->parent;
		*branch = node->right;
	}
	else
		*branch = NULL;

	node->left = node->right = node->parent = NULL;
}


/*
================
list_add_start
================
*/

void list_add_start (tree_list_t *llist, tree_t *item)
{
	assert(llist);
	assert(item);

	if (!llist->nitems) llist->item_end = item;
	if (item->next = llist->item_start) item->next->prev = item;
	llist->item_start = item;
	llist->nitems++;
}

/*
================
list_add_end
================
*/

void list_add_end (tree_list_t *llist, tree_t *item)
{
	assert(llist);
	assert(item);

	if (!llist->nitems)
		llist->item_start = item;
	else
	{
		llist->item_end->next = item;
		item->prev = llist->item_end;
	}
	llist->item_end = item;
	llist->nitems++;
}

/*
==============
list_del_item
==============
*/

void list_del_item (tree_list_t *llist, tree_t *item)
{
	assert(llist);
	assert(item);

	/* We can't delete from an empty list */
	if (!llist->nitems || !llist->item_start) return;

	if (item == llist->item_start)		/* First Item */
	{
		if (llist->item_start = item->next)
			item->next->prev = NULL;
	}
	else if (item == llist->item_end)	/* Last Item */
	{
		if (llist->item_end = item->prev)
			item->prev->next = NULL;
	}
	else
	{
		item->prev->next = item->next;
		item->next->prev = item->prev;
	}

	if (--llist->nitems == 0) llist->item_start = llist->item_end = (void *) 0; /* NULL */
}
