/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: ircd_callbacks.c,v 1.15 2002/01/27 22:34:36 zarjazz Exp $
*/

#include <ircd_include.h>

/* ======================================================== *
 * = These are the functions called by the worker threads = *
 * ======================================================== */

void _client_add (client_t *chan);
void _client_del (client_t *chan);

void _channel_add (channel_t *chan);
void _channel_del (channel_t *chan);

void _server_add (client_t *sv);
void _server_del (client_t *sv);

/*
================
raw_msg
================
*/

#ifdef HAVE_SSL
  #define raw_msg(ssl,sock,msg) _raw_msg(ssl,sock,msg);
#else
  #define raw_msg(ssl,sock,msg) _raw_msg(NULL,sock,msg);
#endif /* HAVE_SSL */

static void _raw_msg (void *ssl, u_int16_t sock, const char *msg)
{
	sys_pkt_t pkt;

	pkt.data = (byte *) msg;
	pkt.size = strlen(msg);

#ifdef HAVE_SSL
	if (ssl)
		ssl_send(&pkt,ssl);
	else
#endif /* HAVE_SSL */
		net_send(&pkt,sock);
}

/*
================
addr_flood_check
================
*/

static bool addr_flood_check (struct sockaddr_in *sa, struct timeval *tv)
{
	u_int32_t ip;
	ip_tree_t *tree;

	/* FIXME: this does not currently detect multiple un-registered clients */
	ip = ntohl(sa->sin_addr.s_addr);
	tree = ip_find(ip);
 
	if (tree && tv->tv_sec)
	{
		long diff = tv->tv_sec - tree->tv_sec;
		/* only allow 1 connection per IP every 5 seconds */
		if (diff < 5) return true;
	}

	return false;
}

/*
================
addr_clone_check
================
*/

static bool addr_clone_check (struct sockaddr_in *sa, unsigned clones)
{
	u_int32_t ip, count;

	ip = ntohl(sa->sin_addr.s_addr);
	count = ip_count(ip,32);
	/* only allow 'clones' connection per ip */
	return (count < clones) ? false : true;
}

/*
================
kline_check
================
*/

static char *kline_check (client_t *cl)
{
	block_t *block = NULL;
	block_t *root = Config;
	block_arg_t *blk_arg;

	char userhost[USERHOSTLEN];	/* nick!user@host of client */
	char userip  [USERIPLEN];	/* nick!user@ip   of client */

	/* generate the userhost */
	client_userhost(cl, userhost, false);
	client_userip(cl, userip, false);

	/* step through Kline blocks */
	while ( block = blocks_by_name(root,block,"Kline") )
	{
		blk_arg = NULL;
		/* match host or ip mask */
		while ( blk_arg = arguments_by_name(block,blk_arg,"Mask") )
		{
			if (str_match(userip,   blk_arg->arg_value, false) ||
				str_match(userhost,	blk_arg->arg_value, false))
			{
				char *blk_val = argument_by_name(block,"Msg");
				if (blk_val) return blk_val;
				return "";
			}
		}
	}
	return NULL;
}

/*
================
raw_accept
----------------
accept a new connection
================
*/

#ifdef HAVE_SSL
static int raw_accept (listener_t *listener, struct sockaddr_in *addr, SSL **ssl_ret)
{
	SSL *ssl = NULL;
#else
static int raw_accept (listener_t *listener, struct sockaddr_in *addr)
{
#endif /* HAVE_SSL */
	int rc, addrlen = sizeof(*addr);
	u_int16_t sock;

	rc = accept(listener->sock, (struct sockaddr *) addr, &addrlen);
	bit_clr(listener,FL_BUSY);

	if (rc < 0) { sys_dprintf(1,"client - accept: %s\n", net_str_error()); return rc; }
	sock = rc;

	/* if SSL listener, do the handshake (in blocking mode) */
#ifdef HAVE_SSL
	if (listener->use_ssl)
	{
		*ssl_ret = ssl = SSL_new(Ctx);
		if (!ssl)
		{
			sys_warn("SSL_new");
			raw_msg(NULL,sock,"NOTICE ERR :SSL Handshake\r\n");
			goto err;
		}

		rc = SSL_set_fd(ssl, sock);
		if (rc != 1)
		{
			sys_warn("SSL_set_fd");
			raw_msg(NULL,sock,"NOTICE ERR :SSL Handshake\r\n");
			goto err ;
		}

		rc = SSL_accept(ssl);
		if (rc != 1)
		{
			unsigned long err;
			char msg[BUFSIZE];

			sys_warn("SSL Client Accept: %d", rc);
			while (err = ERR_get_error()) sys_warn("  %s", ERR_error_string(err, msg));
			raw_msg(NULL,sock,"NOTICE ERR :SSL Handshake\r\n");
			goto err;
		}
	}
#endif /* HAVE_SSL */

	return sock;
err:
#ifdef HAVE_SSL
	if (ssl) { SSL_shutdown(ssl); SSL_free(ssl); }
#endif /* HAVE_SSL */
	net_sock_close(sock);
	return -1;
}

/*
================
callback_client_accept
----------------
accept a new client connection
================
*/

void callback_client_accept (void *arg1, void *arg2)
{
	struct sockaddr_in addr;
	char msg[BUFSIZE], *str = NULL;
	char cl_ip[IPLEN];
	ip_tree_t *ip_tree;
	u_int32_t ip;
	u_int16_t sock;
	client_t *cl;
	int rc;
#ifdef HAVE_SSL
	SSL *ssl = NULL;
#endif /* HAVE_SSL */

	listener_t *listener = (listener_t *) arg1;
	assert(listener);

	/* accept the connection */
#ifdef HAVE_SSL
	rc = raw_accept(listener,&addr,&ssl);
#else
	rc = raw_accept(listener,&addr);
#endif /* HAVE_SSL */
	if (rc < 0) return;
	sock = rc;

	/* check this server can accept new clients */
	if (Options->max_users && UserStats.N_users >= Options->max_users)
		str = "NOTICE AUTH :Max Server Connections -- refused\r\n";

	/* check this listener can accept new clients */
	if (listener->user_max && listener->user_count >= listener->user_max)
		str = "NOTICE AUTH :Max Port Connections -- refused\r\n";

	/* check the incoming ip address is not connection flooding */
	if ( addr_flood_check(&addr, &Now) )
		str = "NOTICE ERR :IP trying to reconnect too fast\r\n";

	/* check for clones */
	if ( Options->max_clones && addr_clone_check(&addr, Options->max_clones) )
		str = "NOTICE ERR :Too many connections from your IP\r\n";

	if (str)
	{
	#ifdef HAVE_SSL
		if (ssl) { SSL_shutdown(ssl); SSL_free(ssl); }
	#endif /* HAVE_SSL */
		net_sock_close(sock);
		return;
	}

	/* create the new client structure */
	cl = sys_malloc(sizeof(*cl));
	cl->tree_key = cl->name;

	bit_set(cl,FL_LOCAL);
	cl->local.server = Srv;

	cl->listener = listener;
	cl->sock = sock;
	cl->sa = addr;

#ifdef HAVE_SSL
	if (listener->use_ssl) cl->ssl = ssl;
#endif /* HAVE_SSL */

#ifdef WINVER
	str_copy(cl_ip, inet_ntoa(addr.sin_addr), sizeof(cl_ip));
#else
	inet_ntop(AF_INET, &addr.sin_addr, cl_ip, sizeof(cl_ip));
#endif /* WINVER */

	sys_dprintf(3,"client connect: %s:%hu\n", cl_ip, ntohs(addr.sin_port));

	/* set initial timestamp of client accept */
	cl->tv_connect = Now.tv_sec;

	/* sendq */
	sendq_init(&cl->sendq);

	/* network welcome */
#ifdef HAVE_SSL
	if (cl->ssl)
	{
		bit_set(cl,FL_SSL);
		snprintf(msg, sizeof(msg), "NOTICE AUTH :*** Using SSL cipher %s\r\n", SSL_get_cipher(cl->ssl));
		raw_msg(ssl,sock,msg);
	}
#endif /* HAVE_SSL */
	snprintf(msg, sizeof(msg), "NOTICE AUTH :*** Welcome to %s\r\n", Srv->name);
	raw_msg(ssl,sock,msg);
	snprintf(msg, sizeof(msg), "NOTICE AUTH :*** Resolving IP %s\r\n", cl_ip);
	raw_msg(ssl,sock,msg);

	READ_LOCK(rwlock_ip);

	/* add client / ip to lists */
	ip = ntohl(addr.sin_addr.s_addr);
	ip_tree = ip_add(ip,Now.tv_sec,false);

	assert(ip_tree);

	/* obtain the hostname of client */
	if (!ip_tree->left)
	{
		/* not cached, so do a dns lookup of the ip address */
		struct hostent *host_ent = gethostbyaddr((char * ) &addr.sin_addr, sizeof(addr.sin_addr), AF_INET);
		if (host_ent)
			ip_tree->left = (void *) str_dup(host_ent->h_name);
		else
			ip_tree->left = (void *) str_dup(cl_ip);
		/* create the encrypted host */
		ip_tree->right = (void *) host_encrypt((const char *) ip_tree->left);
	}
	/* set hostname from cache */
	cl->hostcrypt = (char *) ip_tree->right;
	cl->host	  = (char *) ip_tree->left;
	if ( strcmp(cl->host,cl_ip) )
	{
		bit_set(cl,FL_RESOLVED);
		snprintf(msg, sizeof(msg), "NOTICE AUTH :*** Resolved host %s\r\n", cl->host);
		raw_msg(ssl,sock,msg);
	}
	else
		raw_msg(ssl,sock,"NOTICE AUTH :*** Unable to resolve host\r\n");

	/* are we using encrypted host names? */
	if (Options->encrypt_hosts) bit_set(cl,FL_ENCRYPTHOST);

	/* NOW we set the new client socket non-blocking */
	net_sock_noblock(sock);

	/* what is the domain of the client? */
	cl->domain = strchr(cl->host,'.');
	if (cl->domain)
		cl->domain++;
	else
		cl->domain = cl->host;
	/*
	 * Zarjazz: since we now sort the client list by nick, and this client
	 * hasn't sent a nick yet, we are in a bit of a sticky situation.
	 *
	 * Fix #1 - have a seperate list just for pending clients
	 * Fix #2 - give the client a temporary random nick
	 * Fix #3 - go and have a beer and forget about coding
	 *
	 * now #3 is very tempting but currently we select door #2
	 */
	WRITE_LOCK(rwlock_clients);
	do
	{
		int a,b;
		a = rand(); b = rand();
		memcpy(&cl->name[0],&a,sizeof(int));
		memcpy(&cl->name[4],&b,sizeof(int));
	} while ( client_find(cl->name) );
	_client_add(cl);
	/* check kline */
	if (str = kline_check(cl))
	{
		char err[256];
		snprintf(err,sizeof(err),"K-Lined: %s",str);
		client_exit(cl,err);
	}
	WRITE_UNLOCK(rwlock_clients);
	READ_UNLOCK(rwlock_ip);
}

/*
================
callback_server_accept
----------------
accept a new server connection (this allocated structure is temporary)
================
*/

void callback_server_accept (void *arg1, void *arg2)
{
	struct sockaddr_in addr;
	u_int16_t sock;
	client_t *sv;
	int rc;
#ifdef HAVE_SSL
	SSL *ssl = NULL;
#endif /* HAVE_SSL */

	listener_t *listener = (listener_t *) arg1;
	assert(listener);

	/* accept the connection */
#ifdef HAVE_SSL
	rc = raw_accept(listener,&addr,&ssl);
#else
	rc = raw_accept(listener,&addr);
#endif /* HAVE_SSL */
	if (rc < 0) return;
	sock = rc;

	/* create the new server structure */
	sv = server_malloc(0);

	sv->listener = listener;
	sv->sock = sock;
	sv->sa = addr;

	bit_set(sv,FL_CONNECTED | FL_LOCAL);

#ifdef HAVE_SSL
	if (listener->use_ssl) { sv->ssl = ssl; bit_set(sv,FL_SSL); }
#endif /* HAVE_SSL */

	/* set initial timestamp of client accept */
	sv->tv_connect = Now.tv_sec;

	/* what is the domain of the client? */
	sv->domain = Options->network;

	/* NOW we set the new client socket non-blocking */
	net_sock_noblock(sock);

	/* do the same as in callback_client_accept */
	WRITE_LOCK(rwlock_clients);
	do
	{
		int a,b;
		a = rand(); b = rand();
		memcpy(&sv->name[0],&a,sizeof(int));
		memcpy(&sv->name[4],&b,sizeof(int));
	} while ( client_find(sv->name) );
	_server_add(sv);
	WRITE_UNLOCK(rwlock_clients);
}

/*
================
callback_client_read
----------------
read raw data from a connected client socket
================
*/

void callback_client_read (void *arg1, void *arg2)
{
	u_int32_t id = (u_int32_t) arg1;
	client_t *cl;

	READ_LOCK(rwlock_clients);
	/*
	 * since the client passed to this callback could of been deleted
	 * by another thread before we got the lock we have to check this
	 * client still exists, bummer! oh well thats the price we pay for
	 * trying to be clever and using threads
	 */
	cl = client_validate(id);
	if (cl) client_read(cl);
	READ_UNLOCK(rwlock_clients);
}

/*
================
callback_client_send
----------------
send raw data to a connected client socket
================
*/

void callback_client_send (void *arg1, void *arg2)
{
	u_int32_t id = (u_int32_t) arg1;
	client_t *cl;
	sendq_pkt_t *entry;

	READ_LOCK(rwlock_clients);
	cl = client_validate(id);
	if (cl)
	{
		while (entry = sendq_remove(&cl->sendq))
		{
			int rc;

			if (bit_tst(cl,FL_ZOMBIE))
			{
				sys_free(entry->pkt.data);
				sys_free(entry);
				continue;
			}

		#ifdef HAVE_SSL
			if (cl->ssl)
				rc = ssl_send(&entry->pkt,cl->ssl);
			else
		#endif /* HAVE_SSL */
				rc = net_send(&entry->pkt,cl->sock);

			sys_free(entry->pkt.data);
			sys_free(entry);

			if (rc <= 0) client_exit(cl,"Network Write Error");
			/* try and work around sendq flood on connection */
			if (is_user(cl))
			{
				unsigned len = CLIENT_BUFFER;
				if ((Now.tv_sec - cl->tv_connect) < 60) len <<= 2;
				if (cl->sendq.nbytes > len)
				client_exit(cl,"SendQ Overflow");
			}
		}
		bit_clr(cl,FL_SENDING);
	}
	READ_UNLOCK(rwlock_clients);
}

/*
================
callback_invite_remove
================
*/

void callback_invite_remove (void *arg1, void *arg2)
{
	u_int32_t id = (u_int32_t) arg1;
	index_t *index = (index_t *) arg2;
	client_t *cl;

	WRITE_LOCK(rwlock_clients);
	cl = client_validate(id);
	if (cl)
	{
		ll_del_item(&cl->invite_index,index);
		sys_free(index);
	}
	WRITE_UNLOCK(rwlock_clients);
}

/*
================
callback_ban_remove
================
*/

void callback_ban_remove (void *arg1, void *arg2)
{
	u_int32_t id = (u_int32_t) arg1;
	ban_t *ban = (ban_t *) arg2;
	channel_t *chan;

	WRITE_LOCK(rwlock_channels);
	chan = channel_validate(id);
	if (chan)
	{
		ll_del_item(&chan->bans,ban);
		sys_free(ban);
	}
	WRITE_UNLOCK(rwlock_channels);
}

/*
================
index_cleanup
================
*/

static void index_cleanup (client_t *cl, channel_t *chan, index_t *cl_idx)
{
	index_t *ch_idx = cl_idx->cref;

	/* remove client / channel indexes */
	ll_del_item(&cl->chan_index,cl_idx);
	ll_del_item(&chan->members,ch_idx);

	sys_free(cl_idx);
	sys_free(ch_idx);

	/* remove channel if empty */
	if (!chan->members.nitems) _channel_del(chan);
}

/*
================
callback_client_exit
----------------
remove the client from the network "nicely"
================
*/

void callback_client_exit (void *arg1, void *arg2)
{
	index_t *index, *next;
	client_t *cl;

	u_int32_t id  = (u_int32_t) arg1;
	char *comment = (char *) arg2;
	/*
	 * we have to rw-lock in the correct order here otherwise its
	 * possible to cause the classic thread race
	 */
	WRITE_LOCK(rwlock_clients);
	/* remove client from all channels */
	cl = client_validate(id);
	if (cl)
	{
		WRITE_LOCK(rwlock_channels);
		for (index = (index_t *) cl->chan_index.item_start; index; index = next)
		{
			channel_t *chan;

			next = index->next;
			chan = (channel_t *) index->item;

			/* send the channel quit message */
			channel_part(chan,cl,cl,comment,CE_QUIT);

			/* remove client / channel indexes */
			index_cleanup(cl,chan,index);
		}
		WRITE_UNLOCK(rwlock_channels);
		/* cleanup client invite list */
		for (index = (index_t *) cl->invite_index.item_start; index; index = next)
		{
			next = index->next;
			ll_del_item(&cl->invite_index,index);
			sys_free(index);
		}
		/* tell +s users who disconnected */
		if (Options->luser_connect) server_notice("User Disconnected", cl->name);
		/* remove the client object */
		_client_del(cl);
	}
	WRITE_UNLOCK(rwlock_clients);
	sys_free(comment);
}

/*
================
callback_server_exit
----------------
remove the server from the network "nicely"
================
*/

void callback_server_exit (void *arg1, void *arg2)
{
	client_t *cl;

	u_int32_t id  = (u_int32_t) arg1;
	char *comment = (char *) arg2;
	/*
	 * we have to rw-lock in the correct order here otherwise its
	 * possible to cause the classic thread race
	 */
	WRITE_LOCK(rwlock_clients);
	/* remove client from all channels */
	cl = client_validate(id);
	if (cl)
	{
		/* remove the client object */
		_server_del(cl);
	}
	WRITE_UNLOCK(rwlock_clients);
	sys_free(comment);;
}

/*
================
callback_remove_indexes
----------------
remove a client index from a channel
================
*/

void callback_remove_indexes (void *arg1, void *arg2)
{
	channel_t *chan;
	u_int32_t id	= (u_int32_t)	arg1;
	index_t *index	= (index_t *)	arg2; /* this index is the client index */
	client_t *target= (client_t *)	index->cref->item;
	/*
	 * we have to rw-lock in the correct order here otherwise its
	 * possible to cause the classic thread race
	 */
	WRITE_LOCK(rwlock_clients);
	WRITE_LOCK(rwlock_channels);
	/* remove client index from channel */
	chan = channel_validate(id);
	if (chan) index_cleanup(target,chan,index);
	WRITE_UNLOCK(rwlock_channels);
	WRITE_UNLOCK(rwlock_clients);
}

/*
================
callback_ircd_rehash
----------------
rehash the .conf file & reload the dynamic modules
================
*/

void callback_ircd_rehash (void *arg1, void *arg2)
{
	char userhost[USERHOSTLEN];	/* nick!user@host of client */
	char *who = "SIGHUP signal";
	u_int32_t id  = (u_int32_t) arg1;
	bool sighup = (bool) arg2;
	/*
	 * we have to rw-lock in the correct order here otherwise its
	 * possible to cause the classic thread race
	 */
	WRITE_LOCK(rwlock_clients);
	WRITE_LOCK(rwlock_commands);
	WRITE_LOCK(rwlock_channels);

	/* remove client from all channels */
	if (!sighup)
	{
		client_t *cl = client_validate(id);
		if (cl)
		{
			/* generate the userhost */
			client_userhost(cl,userhost,true);
			who = userhost;
			/* tell client we're rehashing */
			cl_send_numeric(cl, RPL_REHASHING, Srv->name);
		}
	}
	/* notify all '+s' users about rehash */
	server_notice("REHASH activated by", who);

	/* do the config rehash */
	cleanup_block(Config);
	Config = init_block(CFILE);

	/* update the Srv-> elements */
	server_options(Config,Options);

	/* we've reloaded the config now lets reload the plugins */
	load_plugins(false, true);

	WRITE_UNLOCK(rwlock_channels);
	WRITE_UNLOCK(rwlock_commands);
	WRITE_UNLOCK(rwlock_clients);
}

