/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: ircd_channel.c,v 1.8 2002/01/29 18:39:39 zarjazz Exp $
*/

#include <ircd_include.h>

/* ================ */

static tree_t *		_channels;
static channel_t *	_ids[NET_MAX_CHANNELS];

rwlock_t rwlock_channels;

/*
================
get_channels
================
*/

tree_t *get_channels (void) { return _channels; }

/*
================
_channel_del
================
*/

void _channel_del (channel_t *chan)
{
	ban_t *ban;

	/* binary tree */
	tree_del(&_channels,(tree_t *) chan);

	/* id table */
	_ids[chan->id] = NULL;

	/* cleanup bans */
	while (ban = (ban_t *) chan->bans.item_start)
	{
		ll_del_item(&chan->bans,ban);
		sys_free(ban);
	}

	/* user stats */
	UserStats.N_chans--;

	/* cleanup */
	sys_free(chan);
}

/*
================
_channel_add
================
*/

void _channel_add (channel_t *chan)
{
	static u_int32_t id = 1;

	/* binary tree */
	tree_add(&_channels, (tree_t *) chan);

	/* id table */
	while (_ids[id]) { id++; if (id >= NET_MAX_CHANNELS) id = 1; }
	_ids[id] = chan;
	chan->id = id;

	/* user stats */
	UserStats.N_chans++;
}

/*
================
cleanup_channels
================
*/

void __cleanup_channels (channel_t *chan)
{
	if (!chan) return;

	__cleanup_channels(chan->left);
	__cleanup_channels(chan->right);

	_channel_del(chan);
}

static void cleanup_channels (void)
{
	WRITE_LOCK(rwlock_channels);
	__cleanup_channels((channel_t *) _channels);
	WRITE_UNLOCK(rwlock_channels);
	RWLOCK_DESTROY(rwlock_channels);
}

/*
================
init_channels
================
*/

void init_channels (void)
{
	_channels = NULL;
	memset(&_ids[0],0,sizeof(_ids));
	RWLOCK_INIT(rwlock_channels);

	// register for cleanup
	sys_cleanup_register(cleanup_channels);
}

/*
================
channel_validate
----------------
check channel still exists. MUST call with a channel lock
================
*/

channel_t *channel_validate (u_int32_t id) { return _ids[id]; }

/*
================
channel_find
----------------
find channel 'name'
================
*/

channel_t *channel_find (const char *name)
{
	return (channel_t *) tree_find(_channels, (char *) name);
}

/*
================
channel_find_nick
----------------
find client 'nick' on channel
================
*/

index_t *channel_find_nick (channel_t *chan, const char *nick)
{
	index_t *index = NULL;
	client_t *cl = client_find(nick);
	if (cl) index = client_find_chan(cl,chan);
	return index;
}

/*
================
channel_create
================
*/

channel_t *channel_create (const char *name)
{
	channel_t *chan;

	assert(name);

	chan = sys_malloc(sizeof(*chan));

	str_copy(chan->name,name,sizeof(chan->name));
	chan->tv_create = Now.tv_sec;
	chan->tv_modify = Now.tv_sec;
	chan->tree_key = chan->name;
	/*
	 * FIXME: this is a required hack otherwise we would
	 * have to start using write locks in the plugin modules
	 */
	_channel_add(chan);

	return chan;
}

/*
================
channel_mode_str
================
*/

int channel_mode_str (char *buf, char fl, u_int32_t flags, char *key, u_int16_t limit)
{
	char ext_buf[KEYLEN * 2], *ext = ext_buf, *mode = buf;

	*mode++ = fl;

	/* if no flags set */
	if (!flags) { strcpy(mode,"\r\n"); return 3; }

#ifdef HAVE_SSL
	if (flags & CM_SSLONLY)		*mode++ = 'x';
#endif /* HAVE_SSL */
	if (flags & CM_SECRET)		*mode++ = 's';
	if (flags & CM_MODERATED)	*mode++ = 'm';
	if (flags & CM_NOPRIVMSGS)	*mode++ = 'n';
	if (flags & CM_TOPICLIMIT)	*mode++ = 't';
	if (flags & CM_INVITEONLY)	*mode++ = 'i';

	/* modes that require extra arguments */
	*ext = '\0';
	if (flags & CM_KEY)
	{
		*mode++ = 'k';
		if (key[0]) ext += sprintf(ext,"%s",key);
	}
	if (flags & CM_LIMIT)
	{
		*mode++ = 'l';
		if (limit)
		{
			if (ext_buf[0]) *ext++ = ' ';
			ext += sprintf(ext,"%hu",limit);
		}
	}
	if (ext_buf[0])
	{
		int len;

		*mode++ = ' ';
		len = ext - ext_buf;
		memcpy(mode,ext_buf,len);
		mode += len;
	}

	*mode++ = '\r';
	*mode++ = '\n';
	*mode   = '\0';

	return (mode - buf);
}

/*
================
channel_parse_mode_str
================
*/

void channel_parse_mode_str (client_t *cl, channel_t *chan, char **args, bool priv)
{
	char buf[BUFSIZE], userhost[USERHOSTLEN];	/* nick!user@host of client */
	u_int32_t new_flags = chan->flags;
	u_int16_t new_limit = chan->limit;
	char new_key[sizeof(chan->key)];

	bool fl_key = false, fl_limit = false;
	bool fl_not_op = false;

	int val = 0, action = +1;
	char ch, *mode;
	u_int32_t diff;

	new_key[0] = '\0';

	/* generate the userhost */
	client_userhost(cl,userhost,true);

	/* parse the mode string */
	while (mode = args[val++]) while (ch = *mode++)
	{
		index_t *target;
		u_int32_t flag;
		char  *ban_str;
		ban_t *ban;

		switch (ch)
		{
		case '+':
			action = +1;
			break;
		case '-':
			action = -1;
			break;
		/* basic channel modes */
	#ifdef HAVE_SSL
		case 'x':
			if (!bit_tst(cl,FL_SSL)) break;
	#endif /* HAVE_SSL */
		case 's':
		case 'm':
		case 'n':
		case 't':
		case 'i':
			flag = chan_mode(ch);
			if (action > 0)
				new_flags |=  flag;
			else
				new_flags &= ~flag;
			break;
		/* channel modes that require arguments */
		case 'k':
			/* ignore missing arguments */
			if (args[val] == NULL) break;

			/* only allow 1 key change per message */
			if (fl_key) { val++; break; }
			fl_key = true;

			flag = chan_mode(ch);
			if (action > 0)
			{
				str_extract_alnum(new_key,args[val++],sizeof(new_key));
				if (new_key[0]) new_flags |= flag;
			}
			else
			{
				/* FIXME: sanity check current key? */
				val++;
				new_flags &= ~flag;
				new_key[0] = '\0';
			}
			break;
		case 'l':
			flag = chan_mode(ch);
			if (action > 0)
			{
				/* ignore missing arguments */
				if (args[val] == NULL) break;

				if (fl_limit)
					val++;
				else
				{
					new_limit = atoi(args[val++]);
					if (new_limit) new_flags |= flag;
				}
			}
			else if (!fl_limit)
			{
				new_flags &= ~flag;
				new_limit = 0;
			}

			/* only allow 1 limit change per message */
			fl_limit = true;

			break;
		/* user modes */
		case 'o':
		case 'v':
			/* ignore missing arguments */
			if (args[val] == NULL) break;

			/* check client privileges */
			if (!priv) { val++; fl_not_op = true; break; }

			/* check target is on channel */
			target = channel_find_nick(chan,args[val++]);
			if (target)
				target = target->cref;
			else
				{ cl_send_numeric(cl, ERR_NOSUCHNICK, args[val-1]); break; }

			/* apply changes */
			diff = target->flags;
			flag = chan_mode(ch);
			if (action > 0)
				diff |=  flag;
			else
				diff &= ~flag;
			if (target->flags != diff)
			{
				target->flags = diff;
				chan_send_fmt(chan, NULL, FL_BROADCAST | FL_LOCAL, 0,
					":%s MODE %s %c%c %s", userhost, chan->name,
					(action > 0) ? '+' : '-', ch, ((client_t *) target->item)->name);
			}
			break;
		/* bans */
		case 'b':
			if (args[val] == NULL)
			{
				if (is_local(cl))
				{
					/* list bans on channel */
					for (ban = (ban_t *) chan->bans.item_start; ban; ban = ban->next)
						cl_send_numeric(cl, RPL_BANLIST, chan->name, ban->str, ban->who, ban->when);
					cl_send_numeric(cl, RPL_ENDOFBANLIST, chan->name);
				}
				break;
			}

			/* check client privileges */
			if (!priv) { val++; fl_not_op = true; break; }

			/* check the ban is valid */
			ban_str = args[val++];
			ban_str = sanify_ban(ban_str, ban_str);
			if (!ban_str || ban_str[0] == '\0' || strlen(ban_str) > (BANLEN - 1)) break;

			/* add ban */
			if (action > 0)
			{
				ban_add(chan,cl,ban_str,userhost);
			}
			/* del ban */
			else
			{
				/* remove ban if present on channel */
				for (ban = (ban_t *) chan->bans.item_start; ban; ban = ban->next)
					if ( !stricmp(ban->str, ban_str) ) { ban_remove(chan,ban,userhost); break; }
			}

			break;
		/* syntax error */
		default:
			if (is_local(cl)) cl_send_numeric(cl, ERR_UNKNOWNMODE, ch);
			break;
		}
	}

	/* apply any channel mode changes */
	mode = buf;
	/* '-' mode change string */
	diff = ~new_flags &  chan->flags;
	if (diff) mode += channel_mode_str(mode,'-',diff,chan->key,0) - 2; /* strip trailing "\r\n" */
	/* '+' mode change string */
	diff =  new_flags & ~chan->flags;
	if (diff) mode += channel_mode_str(mode,'+',diff,new_key,new_limit);
	/* has anything changed? */
	if (mode != buf)
	{
		/* check client privileges */
		if (!priv)
			fl_not_op = true;
		else
		{
			chan_send_fmt(chan, NULL, FL_BROADCAST | FL_LOCAL, 0,
				":%s MODE %s %s", userhost, chan->name, buf);
			chan->flags = new_flags;
			chan->limit = new_limit;
			strcpy(chan->key,new_key);
			chan->tv_modify = Now.tv_sec;
		}
	}

	/* send warning to local clients if not enough privs */
	if (fl_not_op && is_local(cl)) cl_send_numeric(cl, ERR_CHANOPRIVSNEEDED, chan->name);
}

/*
================
channel_join
----------------
create the client / channel cross-linked indexes
================
*/

channel_t *channel_join (channel_t *chan, char *name, client_t *cl, char *userhost)
{
	index_t *cl_idx, *ch_idx;

	cl_idx = sys_malloc(sizeof(index_t));
	ch_idx = sys_malloc(sizeof(index_t));

	cl_idx->cref = ch_idx;
	ch_idx->cref = cl_idx;

	/* create channel if required and give the client chan-ops */
	if (!chan)
	{
		chan = channel_create(name);
		bit_set(ch_idx,CF_OP);
	}

	cl_idx->item = chan;
	ch_idx->item = cl;

	ll_add_end(&cl->chan_index,cl_idx);
	ll_add_end(&chan->members,ch_idx);

	/* send JOIN to all LOCAL clients on channel */
	chan_send_fmt(chan, NULL, FL_LOCAL, 0, ":%s JOIN :%s", userhost, chan->name);

	sys_dprintf(3,"client %s joined %s\n", cl->name, chan->name);

	return chan;
}

/*
================
channel_part
================
*/

void channel_part (channel_t *chan, client_t *target, client_t *cl, const char *comment, int type)
{
	char userhost[USERHOSTLEN];	/* nick!user@host of client */

	/* generate the source userhost */
	client_userhost(cl, userhost, true);

	/* send the relevant message to channel */
	if (!comment) comment = "";
	switch (type)
	{
	case CE_QUIT:
		chan_send_fmt(chan, NULL, FL_LOCAL, 0, ":%s QUIT :%s", userhost, comment);
		break;
	case CE_PART:
		chan_send_fmt(chan, NULL, FL_LOCAL, 0, ":%s PART %s :%s", userhost, chan->name, comment);
		break;
	case CE_KICK:
		chan_send_fmt(chan, NULL, FL_LOCAL, 0, ":%s KICK %s %s :%s", userhost, chan->name, target->name, comment);
		break;
	default:
		sys_error("Unknown Remove Type");
		break;
	}

	/*
	 * if this is the last client, ZOMBIE the channel (by clearing the channel name)
	 * this should helpfully stop a race occuring if a client cycle's the channel
	 * ie PART #chan / JOIN #chan in same message
	 */
	if (chan->members.nitems <= 1) { chan->name[0] = '\0'; bit_set(chan,FL_ZOMBIE); }
}

/*
================
channel_exit
----------------
remove a client from a channel
================
*/

void channel_exit (index_t *cl_idx, client_t *cl, const char *comment, int type)
{
	index_t   * ch_idx = cl_idx->cref;
	channel_t * chan   = cl_idx->item;
	client_t  * target = ch_idx->item;

	bit_set(cl_idx,FL_ZOMBIE);
	bit_set(ch_idx,FL_ZOMBIE);

	channel_part(chan,target,cl,comment,type);
	/* remove client from channel & channel from client */
	task_create(callback_remove_indexes, (void *) chan->id, cl_idx);
}

/*
================
channel_list
================
*/

static void __channel_list (channel_t *chan, client_t *cl, bool priv, list_options_t *opts)
{
	int chan_mins, topic_mins;

	if (!chan) return;

	chan_mins = (Now.tv_sec - chan->tv_create) / 60;
	topic_mins = chan->tv_topic ? (Now.tv_sec - chan->tv_topic) / 60 : 0;

	/* if secret, check client can see channel */
	if (priv || !bit_tst(chan, CM_SECRET) || client_find_chan(cl,chan))
		if (((signed)   opts->min_users < (signed)   chan->members.nitems) &&
			((unsigned)	opts->max_users > (unsigned) chan->members.nitems) &&

			((signed)   opts->min_chan  < (signed)   chan_mins) &&
			((unsigned) opts->max_chan  > (unsigned) chan_mins) &&

			((signed)   opts->min_topic < (signed)   topic_mins) &&
			((unsigned) opts->max_topic > (unsigned) topic_mins))
				cl_send_numeric(cl, RPL_LIST, chan->name, chan->members.nitems, chan->topic);

	__channel_list(chan->left,  cl, priv, opts);
	__channel_list(chan->right, cl, priv, opts);
}

void channel_list (client_t *cl, bool priv, list_options_t *opts)
{
	__channel_list((channel_t *) _channels, cl, priv, opts);
}

/*
================
channel_names
================
*/

int channel_names (const char *dst, size_t dst_len, channel_t *chan, index_t **marker, bool show)
{
	int rem = dst_len, len = -1;
	index_t *index = *marker;
	char *str = (char *) dst;
	client_t *cl;

	if (!index) index = (index_t *) chan->members.item_start;
	if (!index) goto done;

	for (*str = '\0'; index; index = index->next)
	{
		cl = (client_t *) index->item;

		/* if invisible, check client can see target */
		if (bit_tst(cl,FL_INVISIBLE) && !show) continue;

		len = strlen(cl->name);
		if (len > (rem - 4)) goto done;

		/* write users channel flags */
		if ( bit_tst(index,CF_OP) )
		{
			*str++ = '@';
			rem--;
		}
		else if ( bit_tst(index,CF_VOICE) )
		{
			*str++ = '+';
			rem--;
		}

		/* write users name */
		memcpy(str,cl->name,len);
		str += len;
		rem -= len;

		/* write space */
		*str++ = ' ';
		rem--;
	}
	if (str > dst) --str;
done:
	*marker = index;
	/* write terminator */
	*str++ = '\r';
	*str++ = '\n';
	*str   = '\0';

	return (str - dst);
}
