/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: ircd_client.c,v 1.9 2002/01/27 22:34:36 zarjazz Exp $
*/

#include <ircd_include.h>

/* ================ */

static tree_t *		_clients;
static client_t *	_ids[NET_MAX_CLIENTS];

rwlock_t rwlock_clients;

/*
================
get_channels
================
*/

tree_t *get_clients (void) { return _clients; }

/*
================
_client_del
================
*/

void _client_del (client_t *cl)
{
	index_t *index;
	sendq_pkt_t *entry;

	/* binary tree */
	tree_del(&_clients,(tree_t *) cl);

	/* id table */
	_ids[cl->id] = NULL;

	/* list */
	if (cl->listener) list_del_item(&cl->listener->list, (tree_t *) cl); 

	/* network */
	while (entry = sendq_remove(&cl->sendq))
	{
		sys_free(entry->pkt.data);
		sys_free(entry);
	}

	/* server only cleanup done */
	if (is_server(cl)) return;

	/* local client cleanup */
	if (is_local(cl))
	{
		char ip_port[IPPORTLEN];
		u_int32_t ip;

		/* del ip from list */
		ip = ntohl(cl->sa.sin_addr.s_addr);
		ip_del(ip);

		/* user stats */
		cl->listener->user_count--;

	#ifdef HAVE_SSL
		if (cl->ssl)
		{
			SSL_shutdown(cl->ssl);
			SSL_free(cl->ssl);
		}
	#endif /* HAVE_SSL */
		net_sock_close(cl->sock);
		net_str_sockaddr(ip_port,&cl->sa);
		sys_dprintf(3,"client deleted: %s\n",ip_port);
	}

	/* cleanup invites */
	while (index = (index_t *) cl->invite_index.item_start)
	{
		ll_del_item(&cl->invite_index,index);
		sys_free(index);
	}

	/* user stats */
	UserStats.N_users--;
	if (bit_tst(cl,FL_OPERATOR))  UserStats.N_opers--;
	if (bit_tst(cl,FL_INVISIBLE)) UserStats.N_invis--;

	sys_free(cl);
}

/*
================
_client_add
================
*/

void _client_add (client_t *cl)
{
	static u_int32_t id = 1;

	assert(cl);

	/* binary tree */
	tree_add(&_clients, (tree_t *) cl);

	/* id table */
	while (_ids[id]) { id++; if (id >= NET_MAX_CLIENTS) id = 1; }
	_ids[id] = cl;
	cl->id = id;

	/* user stuff (servers dont apply) */
	if (is_user(cl))
	{
		if (is_local(cl)) cl->listener->user_count++;
		if (++UserStats.N_users > UserStats.N_max_users) UserStats.N_max_users = UserStats.N_users;
	}

	/* list */
	if (cl->listener) list_add_start(&cl->listener->list, (tree_t *) cl); 
}

/*
================
cleanup_clients
================
*/

void __cleanup_clients (client_t *cl)
{
	if (!cl) return;

	__cleanup_clients(cl->left);
	__cleanup_clients(cl->right);

	_client_del(cl);
}

static void cleanup_clients (void)
{
	WRITE_LOCK(rwlock_clients);
	__cleanup_clients((client_t *) _clients);
	WRITE_UNLOCK(rwlock_clients);
	RWLOCK_DESTROY(rwlock_clients);
}

/*
================
init_clients
================
*/

void init_clients (void)
{
	_clients = NULL;
	memset(_ids,0,sizeof(_ids));
	RWLOCK_INIT(rwlock_clients);

	// register for cleanup
	sys_cleanup_register(cleanup_clients);
}

/*
================
client_validate
----------------
check client still exists. MUST call with a client lock
================
*/

client_t *client_validate (u_int32_t id) { return _ids[id]; }

/*
================
client_userhost
----------------
generate the clients "nick!user@host" value
================
*/

void client_userhost (client_t *cl, char *dst, bool hide)
{
	if (hide)
	{
		if (Options->encrypt_hosts && bit_tst(cl,FL_ENCRYPTHOST))
		{
			sprintf(dst, "%s!%s@%s.%s", cl->name, cl->user, cl->hostcrypt, 
				(Options->hide_domain || !bit_tst(cl,FL_RESOLVED)) ? Options->network : cl->domain);
			return;
		}
		if (Options->hide_domain)
		{
			sprintf(dst, "%s!%s@%s.%s", cl->name, cl->user, cl->hostcrypt, Options->network);
			return;
		}
	}
	sprintf(dst, "%s!%s@%s", cl->name, cl->user, cl->host);
}

/*
================
client_userip
----------------
generate the clients "nick!user@ip" value
================
*/

void client_userip (client_t *cl, char *dst, bool hide)
{
	if (hide && Options->hide_ip)
		sprintf(dst, "%s!%s@%s", cl->name, cl->user, "127.0.0.1");
	else
	{
		char cl_ip[IPLEN];
	#ifdef WINVER
		str_copy(cl_ip, inet_ntoa(cl->sa.sin_addr), sizeof(cl_ip));
	#else
		inet_ntop(AF_INET, &cl->sa.sin_addr, cl_ip, sizeof(cl_ip));
	#endif /* WINVER */
		sprintf(dst, "%s!%s@%s", cl->name, cl->user, cl_ip);
	}
}

/*
================
client_ping
----------------
start the ping keap-alive
================
*/

void client_ping (client_t *cl)
{
	cl->tv_ping = Now.tv_sec;
	bit_set(cl,FL_PING_SENT);
	cl_send_fmt(cl, "PING :%s", Srv->name);
}

/*
================
client_register
----------------
register client as valid
================
*/

void client_register (client_t *cl)
{
	/* send the connection message */
	cl_send_numeric(cl, RPL_WELCOME_IRC,	cl->name);
	cl_send_numeric(cl, RPL_WELCOME_HOST,	Srv->name, VERSION);
	cl_send_numeric(cl, RPL_WELCOME_SERV,	CDATE, CTIME);
	cl_send_numeric(cl, RPL_WELCOME_VER,	Srv->name, VERSION, "");
	/* numeric 005 is supported since mirc 5.91 */
	cl_send_numeric(cl, RPL_NETWORK_IS,		Options->network);
	/* set flag */
	bit_set(cl,FL_REGISTERED);
	/* tell +s users who connected */
	if (Options->luser_connect) server_notice("User Connected", cl->name);
	/* store the ip address in union to send over network */
//	cl->u.ip = cl->sa.sin_addr.s_addr;
	/* reset pass union */
	cl->u.pass[0] = '\0';

	/* FIXME: notify network of new client */
	{
		char userhost[USERHOSTLEN];	/* nick!user@host of client */
		char umode[32];

		client_userhost(cl,userhost,true); /* hide host details from remote servers */
		client_mode_str(umode,'+',cl->flags);

		cl_send_fmt(NULL, "USER %s %hu %s :%s", userhost, local_server_id(cl), umode, cl->info);
	}
}

/*
================
client_find
----------------
find client 'nick'
================
*/

client_t *client_find (const char *nick)
{
	return (client_t *) tree_find(_clients, (char *) nick);
}

/*
================
client_find_invite
================
*/

index_t *client_find_invite (client_t *cl, channel_t *chan)
{
	index_t *index;

	for (index = (index_t *) cl->invite_index.item_start; index; index = index->next)
		if (chan == index->item) return index;

	return NULL;
}

/*
================
client_find_chan
================
*/

index_t *client_find_chan (client_t *cl, channel_t *chan)
{
	index_t *index;

	for (index = (index_t *) cl->chan_index.item_start; index; index = index->next)
	{
		if (bit_tst(index,FL_ZOMBIE)) continue;
		if (chan == index->item) return index;
	}

	return NULL;
}

/*
================
client_find_chan
================
*/

client_t *client_create_remote (client_t *sv, char *name, char *user, char *host)
{
	client_t *cl;

	cl = sys_malloc(sizeof(*cl));
	cl->tree_key = cl->name;

	strcpy(cl->name,name);
	strcpy(cl->user,user);

	cl->local.server = sv;
	cl->listener = Remote;

	/* set hostname (FIXME: crypt stuff) */
	cl->host = strdup(host);
	if (!isdigit(cl->host[0])) bit_set(cl,FL_RESOLVED);

	/* what is the domain of the client? */
	cl->domain = strchr(cl->host,'.');
	if (cl->domain)
		cl->domain++;
	else
		cl->domain = cl->host;

	/* add client to list(s) */
	_client_add(cl);

	sys_dprintf(3,"remote client: %s\n", cl->name);

	return cl;
}


/*
================
client_select
----------------
select on local client sockets
================
*/

int client_select (listener_t *listener, fd_set *fdp, int maxSock)
{
	client_t *cl;

	READ_LOCK(rwlock_clients);
	for (cl = (client_t *) listener->list.item_start; cl; cl = cl->next)
	{
		/* select on client */
		if (cl->sock)
		{
			/* timeout (PING / PONG) check */
			if ( bit_tst(cl,FL_PING_SENT) )
			{
				int32_t timeout = Options->ping_freq;
				if ( !bit_tst(cl,FL_REGISTERED) ) timeout >>= 2;
				if ( (Now.tv_sec - cl->tv_ping) > timeout ) client_exit(cl,"Ping Timeout");
			}
			else if ( bit_tst(cl,FL_REGISTERED) )
			{
				if ( (Now.tv_sec - cl->tv_ping) > Options->ping_freq )
					client_ping(cl);
			}
			else if ( (Now.tv_sec - cl->tv_connect) > (Options->ping_freq >> 1))
				client_exit(cl,NULL);

			/* select on live socket */
			if ( !bit_tst(cl,FL_BUSY | FL_ZOMBIE) )
			{
				SELECT(cl->sock);
			}
		}
	}
	READ_UNLOCK(rwlock_clients);

	return maxSock;
}

/*
================
client_poll
----------------
detect data on client sockets
================
*/

void client_poll (listener_t *listener, fd_set *fdp)
{
	client_t *cl;

	READ_LOCK(rwlock_clients);
	for (cl = (client_t *) listener->list.item_start; cl; cl = cl->next)
	{
		if (cl->sock && !bit_tst(cl,FL_BUSY | FL_ZOMBIE))
		{
			if (LISTEN(cl->sock))
			{
				bit_set(cl,FL_BUSY);
				task_create(callback_client_read, (void *) cl->id, NULL);
			}
		}
	}
	READ_UNLOCK(rwlock_clients);
}

/*
================
client_accept
================
*/

void client_accept (listener_t *listener)
{
	bit_set(listener,FL_BUSY);
	if (bit_tst(listener,FL_SERVER))
		task_create(callback_server_accept, listener, NULL);
	else
		task_create(callback_client_accept, listener, NULL);
}

/*
================
client_read
----------------
read raw data from a connected client socket
================
*/

void client_read (client_t *cl)
{
	byte *s, *str, *buffer;
	sys_pkt_t pkt;
	char *err = NULL;
	int buf_size;

	buf_size = is_user(cl) ? CLIENT_BUFFER : SERVER_BUFFER;
	buffer = alloca(buf_size);

	pkt.data = buffer;
	pkt.size = 0;

	/* SSL check */
#ifdef HAVE_SSL
	if (cl->ssl)
		ssl_read(&pkt, cl->ssl, (struct sockaddr *) &cl->sa, buf_size);
	else
#endif /* HAVE_SSL */
		net_read(&pkt, cl->sock, (struct sockaddr *) &cl->sa, buf_size);

	if (pkt.size <= 0)
		{ err = "Network Read Error"; goto done; }
	/* client sent way too much data (flood?) */
	else if (pkt.size >= buf_size)
		{ err = "ReadQ Overflow"; goto done; }

	/* ensure buffer is NULL terminated for the parse routine(s) */
	pkt.data[pkt.size] = '\0';
	/* parse & execute valid commands */
	READ_LOCK(rwlock_commands);
	for (str = pkt.data; str && *str; str = s)
	{
		size_t size;

		s = strchr(str,'\n');
		if (s)
			*s++ = '\0';
		else
			s = pkt.data + pkt.size;
		size = s - str;
		if (size >= BUFSIZE) { err = "ReadQ Flood"; break; }
		if (size) cmd_parse(cl,str,size);
	}
	READ_UNLOCK(rwlock_commands);
done:
	if (err)
		client_exit(cl,err);
	else
		bit_clr(cl,FL_BUSY);
}

/*
================
client_exit
----------------
remove a client from the network "nicely"
================
*/

void client_exit (client_t *cl, const char *comment)
{
	if (bit_tst(cl,FL_ZOMBIE)) return;
	bit_set(cl,FL_ZOMBIE);

	if (is_server(cl))
	{
		sys_dprintf(2, "server_exit: %s\n", comment ? comment: "");
		/* note: we have to malloc the comment since it could be a stack variable */
		task_create(callback_server_exit, (void *) cl->id, (void *) (comment ? str_dup(comment) : NULL));
	}
	else
	{
		sys_dprintf(2, "client_exit: %s\n", comment ? comment: "");
		/* note: we have to malloc the comment since it could be a stack variable */
		task_create(callback_client_exit, (void *) cl->id, (void *) (comment ? str_dup(comment) : NULL));
	}
}

/*
================
callback_set_nick
================
*/

void callback_set_nick (void *arg1, void *arg2)
{
	char userhost[USERHOSTLEN];	/* nick!user@host of client */
	index_t *index;
	client_t *cl;

	u_int32_t id	= (u_int32_t) arg1;
	char *nick		= (char *) arg2;
	/*
	 * we have to rw-lock in the correct order here otherwise its
	 * possible to cause the classic thread race
	 */
	WRITE_LOCK(rwlock_clients);

	cl = client_validate(id);
	if (cl && !bit_tst(cl,FL_ZOMBIE))
	{
		/* generate the (current) userhost */
		client_userhost(cl,userhost,true);

		/* update the nick and relink the binary tree */
		tree_del(&_clients,(tree_t *) cl);
		str_copy(cl->name, nick, sizeof(cl->name));
		tree_add(&_clients,(tree_t *) cl);

		/* have we registered yet? */
		if (bit_tst(cl,FL_REGISTERED))
		{
			/* send nick change to client */
			cl_send_fmt(cl, ":%s NICK :%s", userhost, nick);

			/* send nick change to all other users on the clients channel list */
			READ_LOCK(rwlock_channels);
			for (index = (index_t *) cl->chan_index.item_start; index; index = index->next)
			{
				channel_t *chan;

				chan = (channel_t *) index->item;
				if (!bit_tst(chan,FL_ZOMBIE)) chan_send_fmt(chan, cl, 0, 0, ":%s NICK :%s", userhost, nick);
			}
			READ_UNLOCK(rwlock_channels);
		}
	}
	WRITE_UNLOCK(rwlock_clients);
	sys_free(nick);
}

/*
================
client_set_nick
----------------
rename the client
================
*/

void client_set_nick (client_t *cl, const char *nick)
{
	task_create(callback_set_nick, (void *) cl->id, str_dup(nick));
}

/*
================
client_idle_time
----------------
part of whois response (only for local clients)
================
*/

void client_idle_time (client_t *cl, client_t *target)
{
	if (is_local(cl)) cl_send_numeric(cl, RPL_WHOISIDLE, target->name,
		Now.tv_sec - target->tv_active, target->tv_connect);
}
/*
================
client_cookie
================
*/

void client_cookie (client_t *cl)
{
	if (bit_tst(cl,FL_REGISTERED)) return;

	/* send PING as required */
	if (!cl->cookie && (bit_tst(cl,FL_GOTNICK) || cl->user[0]))
	{
		while (!cl->cookie) cl->cookie = rand() & 0x7FFFFFFF;
		cl_send_fmt(cl, "PING :%u", cl->cookie);
		cl->tv_ping = Now.tv_sec;
		bit_set(cl,FL_PING_SENT);
	}
	/* else, register as required */
	else if (cl->cookie == COOKIE_VERIFIED)
	{
		client_register(cl);
		/* send the MOTD */
		show_motd(cl);
	}
}

/*
================
client_pong
================
*/

int client_pong (client_t *cl, char *arg)
{
	/* respond to standard pongs */
	if (is_server(cl) || bit_tst(cl,FL_REGISTERED))
	{
		/* the client is just responding to a PING request */
		cl->tv_ping = Now.tv_sec;
	}
	else
	{
		u_int32_t cookie;

		if (!arg) { cl_send_numeric(cl, ERR_NEEDMOREPARAMS, "PONG"); return 1; }
		cookie = atoi(arg);

		if (cookie != cl->cookie)
		{
			cl_send_fmt(cl,"ERROR: Invalid Cookie %s", arg);
			return -1;
		}
		cl->cookie = COOKIE_VERIFIED;
		if (bit_tst(cl,FL_GOTNICK) && cl->user[0])
		{
			client_register(cl);
			/* send the MOTD */
#ifndef _DEBUG
			show_motd(cl);
#endif
		}
	}

	return 0;
}

/*
================
client_mode_str
================
*/

int client_mode_str (char *buf, char fl, u_int32_t flags)
{
	char *mode = buf;

	*mode++ = fl;

	/* if no flags set */
	if (!flags) { strcpy(mode,"\r\n"); return 3; }

	if (flags & FL_SERVICE)		*mode++ = 'k';
	if (flags & FL_SNOTICE)		*mode++ = 's';
	if (flags & FL_OPERATOR)	*mode++ = 'o';
	if (flags & FL_INVISIBLE)	*mode++ = 'i';
	if (flags & FL_WALLOPS)		*mode++ = 'w';

//	*mode++ = '\r';
//	*mode++ = '\n';
	*mode   = '\0';

	return (mode - buf);
}

/*
================
client_parse_mode_str
================
*/

void client_parse_mode_str (client_t *cl, client_t *target, char *mode, bool priv)
{
	char userhost[USERHOSTLEN];	/* nick!user@host of client */
	char buf[BUFSIZE];
	u_int32_t diff, new_flags = target->flags;
	char ch = *mode;
	int action;
	
	/* generate the userhost */
	client_userhost(cl,userhost,true);
	
	/* parse the mode string */
	while (ch = *mode++)
	{
		u_int32_t flag;
		bool chk;
		
		chk = priv;
		
		switch (ch)
		{
		case '+':
			action = +1;
			break;
		case '-':
			action = -1;
			break;
			/* basic channel modes */
		case 'k':
			/* ignore channel service mode changes */
			break;
		case 'w':
			chk = chk || Options->luser_wallops;
		case 'o':
			if (action > 0 && !chk) break;
		case 's':
		case 'i':
			flag = user_mode(ch);
			if (action > 0)
				new_flags |=  flag;
			else
				new_flags &= ~flag;
			break;
			/* syntax error */
		default:
			cl_send_numeric(cl, ERR_UMODEUNKNOWNFLAG, ch);
			break;
		}
	}
	
	/* apply any user mode changes */
	mode = buf;
	/* '-' mode change string */
	diff = ~new_flags &  target->flags;
	if (diff)
	{
		mode += client_mode_str(mode,'-',diff) - 2; /* strip trailing "\r\n" */
		/* user stats */
		if (diff & FL_INVISIBLE) UserStats.N_invis--;
		if (diff & FL_OPERATOR) UserStats.N_opers--;
	}
	/* '+' mode change string */
	diff =  new_flags & ~target->flags;
	if (diff)
	{
		mode += client_mode_str(mode,'+',diff);
		/* user stats */
		if (diff & FL_INVISIBLE) UserStats.N_invis++;
	}
	/* has anything changed? */
	if (mode != buf)
	{
		target->flags = new_flags;
		cl_send_fmt(cl, ":%s MODE %s :%s", userhost, target->name, buf);
		if (cl != target) cl_send_fmt(target, ":%s MODE %s :%s", userhost, target->name, buf);
	}
}
/*
================
all_send_fmt
================
*/

static void __all_send_fmt (tree_t *tree, u_int32_t filter, byte *buffer, int len)
{
	client_t *cl = (client_t *) tree;

	if (!cl) return;

	if (is_user(cl) && (!filter || bit_tst(cl,filter))) cl_send_raw(cl, buffer, len);

	__all_send_fmt(tree->left, filter, buffer, len);
	__all_send_fmt(tree->right, filter, buffer, len);
}

void all_send_fmt (u_int32_t filter, const char *fmt, ...)
{
	byte buffer[BUFSIZE];
	va_list	argptr;
	int len;

	/* prepare string */
	va_start(argptr,fmt);
	vsnprintf(buffer, sizeof(buffer) - 2, fmt, argptr);
	va_end(argptr);

	/* remove spurious white spaces */
	len = strlen(buffer);
	while (len > 0 && isspace(buffer[len - 1])) buffer[--len] = '\0';

	/* append terminating newline */
	buffer[len++] = '\r';
	buffer[len++] = '\n';
	buffer[len] = '\0';

	/* debug broadcasts */
	sys_dprintf(1,"%s", buffer);

	/* send line to all (FIXME: local) clients */
	__all_send_fmt(_clients, filter, buffer, len);
}
