/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: ircd_config.c,v 1.4 2002/01/13 21:33:36 zarjazz Exp $
*/

#include <ircd_include.h>

/* Globals */

block_t *Config = NULL;

/* ================ */

static int _block_table_size = 0;
static conf_block_t **_block_table = NULL;

/* ================ */

static conf_entry_t _root_entries[] =
{
	{ "ServerName",		CF_REQUIRED },
	{ "ServerID",		CF_REQUIRED },
	{ "Description",	CF_REQUIRED },
	{ "Network",		CF_REQUIRED },
	{ "Workers",		0 },
	{ "PluginDir",		CF_REQUIRED | CF_MULTIREF },
	{ "Motd",			0 },
	{ "ClientBuffer",	0 },
	{ "ServerBuffer",	0 },
	{ "MaxUsers",		0 },
	{ "MaxClones",		0 },
	{ "MaxChannels",	0 },
	{ "MaxBans",		0 },
	{ "PingFreq",		0 },
	{ "ClientHosts",	CF_REQUIRED },
	{ "LuserWallops",	0 },
	{ "LuserConnect",	0 },
#ifdef HAVE_SSL
	{ "SSL",			0 },
	{ "SSLCertificate",	CF_REQUIRED },
#endif /* HAVE_SSL */
	/* Terminator */
	{ NULL, 0 }
};

static conf_block_t _root = { "Root", _root_entries };

/* ================ */

static conf_entry_t	_admin_entries[] = 
{
	{ "Line1",	CF_REQUIRED },
	{ "Line2",	CF_REQUIRED },
	{ "Line3",	CF_REQUIRED },
	/* Terminator */
	{ NULL, 0 }
};

static conf_block_t _admin = { "Admin", _admin_entries };

/* ================ */

static conf_entry_t _listen_entries[] = 
{
	{ "Port",		CF_REQUIRED },
	{ "Type",		CF_REQUIRED },
	{ "MaxUsers",	0 },
#ifdef HAVE_SSL
	{ "SSL",		0 },
#endif /* HAVE_SSL */
	/* Terminator */
	{ NULL, 0 }
};

static conf_block_t _listen = { "Listen", _listen_entries };

/* ================ */

static conf_entry_t _server_entries[] = 
{
	{ "ServerAddress",	CF_REQUIRED },
	{ "Password",		CF_REQUIRED },
#ifdef HAVE_SSL
	{ "SSL",			0 },
#endif /* HAVE_SSL */
	/* Terminator */
	{ NULL, 0 }
};

static conf_block_t _server = { "Server", _server_entries };

/* ================ */

static conf_entry_t	_oper_entries[] =
{
	{ "Password",	CF_REQUIRED },
	{ "Mask",		CF_REQUIRED | CF_MULTIREF },
	{ "AutoMode",	0 },
	/* Terminator */
	{ NULL, 0 }
};

static conf_block_t _oper = { "Oper", _oper_entries };

/* ================ */

static conf_entry_t	_kline_entries[] = 
{
	{ "Mask",	CF_REQUIRED | CF_MULTIREF },
	{ "Msg",	CF_REQUIRED | CF_MULTIREF },
	/* Terminator */
	{ NULL, 0 }
};

static conf_block_t _kline = { "Kline", _kline_entries };

/* ================ */

static conf_entry_t	_reserve_entries[] = 
{
	{ "Nicks",	CF_MULTIREF },
	/* Terminator */
	{ NULL, 0 }
};

static conf_block_t _reserve = { "Reserve", _reserve_entries };

/*
================
cleanup_ircd_config
================
*/

static void cleanup_ircd_config (void)
{
	sys_free(_block_table);
	cleanup_block(Config);
}

/*
================
init_ircd_config
================
*/

block_t *init_ircd_config (const char *fname)
{
	block_t *block;

	// register config blocks
	ircd_config_register(&_root);
	ircd_config_register(&_admin);
	ircd_config_register(&_listen);
	ircd_config_register(&_oper);
	ircd_config_register(&_kline);
	ircd_config_register(&_reserve);
	ircd_config_register(&_server);

	// read config file
	block = init_block(fname);

	// register for cleanup
	sys_cleanup_register(cleanup_ircd_config);

	return block;
}

/*
================
ircd_config_register
================
*/

void ircd_config_register (conf_block_t *block)
{
	_block_table = sys_realloc(_block_table, sizeof(conf_block_t *) * (_block_table_size + 1));
	_block_table[_block_table_size++] = block;
}

/*
================
ircd_config_string
----------------
returns the first match of argument 'arg' inside <block {block_value}> where 'block_value' is optional
if block is NULL, search the <Root> block
================
*/

char *ircd_config_string (block_t *root, const char *block_name, const char *block_value, const char *arg)
{
	block_t *block = NULL;
	char *blk_val;

	assert(arg);

	/* <block> */
	if (block) while ( block = blocks_by_name(root,block,block_name) )
	{
		if (block_value)
		{
			if (!block->block_value) continue;
			if (stricmp(block_value, block->block_value)) continue;
		}

		blk_val = argument_by_name(block,arg);
		if (blk_val) return blk_val;
	}
	/* <Root> */
	else
	{
		blk_val = argument_by_name(root,arg);
		if (blk_val) return blk_val;
	}

	return NULL;
}

/*
================
config_string
----------------
used by plugin interfaces
================
*/

char *config_string (const char *block, const char *block_value, const char *arg)
{
	return ircd_config_string(Config, block, block_value, arg);
}

/*
================
check_mask
================
*/

static void check_masks (block_t *root, const char *name)
{
	block_t *block = NULL;
	block_arg_t *blk_arg;
	char *mask, ban[BANLEN];

	/* sanity check the Mask values */
	while ( block = blocks_by_name(root,block,name) )
	{
		blk_arg = NULL;
		while ( blk_arg = arguments_by_name(block,blk_arg,"Mask") )
		{
			if ( !stricmp(blk_arg->arg_value, "UNRESOLVED") ) continue;

			mask = sanify_ban(ban, blk_arg->arg_value);
			if (!mask || !mask[0])
			{
				if (block->block_value)
					sys_error("Bad Mask in <%s %s>: Mask \"%s\"",
						block->block_name, block->block_value, blk_arg->arg_value);
				else
					sys_error("Bad Mask in <%s>: Mask \"%s\"",
						block->block_name, blk_arg->arg_value);
			}
			strcpy(blk_arg->arg_value,ban);
		}
	}
}

/*
================
ircd_config_check
----------------
sanity checks the supplied config file using the syntax + specific rules
================
*/

void ircd_config_check (block_t *root)
{
	/* syntax check & argument type check */
	block_syntax_check(root,_block_table,_block_table_size);

	/* sanity check the Mask values */
	check_masks(root, "Oper");
	check_masks(root, "Kline");
}
