/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: ircd_main.c,v 1.6 2002/01/16 16:16:12 zarjazz Exp $
*/

#include <ircd_include.h>

/* ================ */

#ifdef WINVER
  #include <process.h>
  typedef int pid_t;
#endif

/* Globals */

#define PROCESS "sircd"

char *VERSION  = "0.4.0 Alpha";
char *CDATE = __DATE__;
char *CTIME = __TIME__;
char *CFILE = PROCESS ".conf";
char *LFILE = PROCESS ".log";
char *PFILE = PROCESS ".pid";

/*
================
write_pid_file
================
*/

static void write_pid_file (const char *fname, pid_t pid)
{
	assert(fname);

	if (pid)
	{
		FILE *f = fopen(fname,"w");
		if (!f) { sys_warn("Unable to open: %s",fname); return; }
		fprintf(f,"%u\n",pid);
		fclose(f);
	}
	else
		remove(fname);
}

/*
===============
write_header
===============
*/

static void write_header (void)
{
 	/* Print Header */
	sys_color(s_red);
	sys_printf("=====================================\n");
	sys_color(s_green);
	sys_printf(" %s: ", PROCESS);
	sys_color(s_cyan);
	sys_printf(" v%s\n", VERSION);
	sys_color(s_green);
	sys_printf(" Author(s)\n");
	sys_color(s_normal);
	sys_printf("   Zarjazz (zarjazz@barrysworld.com)\n");
	sys_color(s_red);
	sys_printf("=====================================\n");
	sys_color(s_normal);
}

/*
================
cleanup_main
================
*/

static void cleanup_main (void)
{
	write_pid_file(PFILE,0);

	sys_color(s_green);
	sys_printf(PROCESS " shutdown\n");
	sys_color(s_normal);
}

/*
===============
init_main
===============
*/

void init_main (void)
{
#if !defined(WINVER) && defined(IRC_UID) && defined(IRC_GID)
	int rc;
	uid_t uid, euid;

	/* UID settings */
	uid = getuid(); euid = geteuid();

	if (uid != euid) sys_error("Please do not run %s with setuid", PROCESS);
	if (IRC_UID < 100 || IRC_GID < 0) sys_error("IRC_(U|G)ID < 100");

	rc = setuid(IRC_UID);
	if (rc < 0) sys_error("setuid() failed");

	rc = setgid(IRC_GID);
	if (rc < 0) sys_error("setgid() failed");
#endif

	// register for cleanup
	sys_cleanup_register(cleanup_main);

	/* init Now */
	gettimeofday(&Now,NULL);

	/* init rand() */
#ifdef WINVER
	srand( (unsigned) time( NULL ) );
#else
	srand(Now.tv_sec);
#endif /* WINVER */

	/* print text banner */
	write_header();
}

/*
================
cleanup
================
*/

void cleanup (void)
{
	static bool done = false;
	if (done) return;
	done = true;

	sys_printf("cleanup:\n");
	/* Call all registered cleanup routines */
	sys_cleanup();
}

/*
================
signal stuff
----------------
pthread signal cleanup is a bitch
================
*/

volatile bool terminate = false;

void sig_handler (int i)
{
	sys_dprintf(1,"signal: %d\n",i);
	terminate = true;
}

void sig_hup (int i)
{
	/* received HUP signal */
	ircd_rehash(NULL);
}

/*
================
main
================
*/

int main (int argc, char *argv[])
{
	int i, workers;
	char *str;

	/* setup signal handler(s) */
#ifndef WINVER
	struct sigaction act;

	memset(&act,0,sizeof(act));
	act.sa_handler = sig_handler;
	sigaction(SIGINT, &act, NULL);

	memset(&act,0,sizeof(act));
	act.sa_handler = sig_hup;
	sigaction(SIGHUP, &act, NULL);

	memset(&act,0,sizeof(act));
	act.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &act, NULL);
#endif

	/* Parse the command line */
	for (i = 1; i < argc; i++)
	{
		int n;
		char *a = argv[i];

		if (*a++ != '-') sys_error("argument: %s",argv[i]);

		switch (*a)
		{
		case 'f':
			CFILE = argv[++i];
			break;
		case 'o':
			LFILE = argv[++i];
			break;
		case 'p':
			PFILE = argv[++i];
			break;
		case 'v':
			for (n = 0; *a == 'v'; a++) n++;
			sys_debug_set(n);
			break;
		default:
			sys_error("argument: %s",argv[i]);
			break;
		}
	}

	/* write parent process '.pid' file */
	write_pid_file(PFILE, getpid());

	/* system init stuff */
	sys_log_init(1, LFILE);

	atexit(&cleanup);

	init_main();
	init_net();
	init_threads();
	init_ip_cache();
	init_numerics();

	/* system init complete */
	sys_color(s_green);
	sys_printf(PROCESS " initialized\n");
	sys_color(s_normal);

	/* parse config */
	Config = init_ircd_config(CFILE);
	if (!Config) sys_error("config file error");
	if ( sys_debug_get() ) print_block(Config,0);
	ircd_config_check(Config);

	/* SSL init stuff */
#ifdef HAVE_SSL
	init_ssl();
#endif /* HAVE_SSL */

	/* server init stuff */
	init_channels();
	init_clients();
	init_local();
	init_plugins();

	/* create threads */
	str = ircd_config_string(Config, NULL, NULL, "Workers");
	workers = str ? atoi(str) : 0;
	if (!workers) workers = 4;

	create_workers(workers);
	create_listeners(Config);
	create_servers(Config);

	/* begin */
	net_loop_parent();

	exit(0);
}
