/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: ircd_misc.c,v 1.4 2002/01/27 22:34:36 zarjazz Exp $
*/

#include <ircd_include.h>

/* ================ */

/* stats */
struct userstats_s UserStats;

/*
================
ircd_terminate
================
*/

void ircd_terminate (void) { terminate = true; }

/*
================
ircd_rehash
================
*/

void ircd_rehash (client_t *cl)
{
	task_create(callback_ircd_rehash, (void *) (cl ? (void *) cl->id : NULL), (void *) (cl ? false : true));
}

/*
================
invite_remove
================
*/

void invite_remove (client_t *cl, index_t *index) { task_create(callback_invite_remove, (void *) cl->id, index); }

/*
================
ban_remove
================
*/

void ban_remove (channel_t *chan, ban_t *ban, char *userhost)
{
	chan_send_fmt(chan, NULL, FL_BROADCAST, 0,
		":%s MODE %s -b %s", userhost, chan->name, ban->str);
	task_create(callback_ban_remove, (void *) chan->id, ban);
}

/*
================
ban_add
---------------
if cl == NULL, then userhost is the source server name
================
*/

void ban_add (channel_t *chan, client_t *cl, char *ban_str, char *userhost)
{
	bool found = false;
	bool removed = false;
	ban_t *ban;

	/* check existing channel bans */
	for (ban = (ban_t *) chan->bans.item_start; ban; ban = ban->next)
	{
		/* does the ban already exist */
		if ( !stricmp(ban->str, ban_str) ) { found = true; break; }
		/* remove overlapping bans */
		if ( str_match(ban->str, ban_str, false) )
		{
			removed = true;
			ban_remove(chan,ban,userhost);
		}
	}

	/* add ban to channel list */
	if (!found)
	{
		if (Options->max_bans && chan->bans.nitems >= Options->max_bans && !removed)
		{
			if (cl) cl_send_numeric(cl,ERR_BANLISTFULL,chan->name,ban_str);
		}
		else
		{
			chan_send_fmt(chan, NULL, FL_BROADCAST, 0,
				":%s MODE %s +b %s", userhost, chan->name, ban_str);
			ban = sys_malloc(sizeof(ban_t));
			strcpy(ban->str, ban_str);
			strcpy(ban->who, cl ? cl->name : userhost);
			ban->when = Now.tv_sec;
			ll_add_end(&chan->bans,ban);
		}
	}
}

/*
================
valid_nick_char
----------------
return 1 if the char is valid for use as a nick
================
*/

int valid_nick_char (int ch)
{
	if (ch >= 'A' && ch <= '}') return 1;
	if (ch >= '0' && ch <= '9') return 1;
	if (ch == '_' || ch == '-') return 1;
	return 0;
}

/*
================
valid_chan_char
----------------
return 1 if the char is valid for use as a channel names
================
*/

int valid_chan_char (int ch)
{
	if (isspace(ch)) return 0;
	if (ch == ',' || ch == '\7' || ch == '\0' ) return 0; /* comma, BELL or NUL */
	return 1;
}

/*
================
sanify_ban
================
*/

char *sanify_ban (char *dst_str, char *ban_str)
{
	char ch, pr = '\0', *dst, *src;
	bool fl_ex, fl_at;

	fl_ex = fl_at = false;
	dst = dst_str;
	src = ban_str;

	while (ch = *src++)
	{
		switch (ch)
		{
		case '*':
			if (pr == '*') continue;
			break;
		case '!':
			/* ! before @, only 1 ! valid */
			if (fl_at || fl_ex) return NULL;
			fl_ex = true;
			break;
		case '@':
			/* min 1 char betwen ! & @, ! before @, only 1 @ valid */
			if (pr == '!' || !fl_ex || fl_at) return NULL;
			fl_at = true;
			break;
		default:
			if (!valid_nick_char(ch) && ch != '.') continue;
			break;
		}
		*dst++ = ch;
		pr = ch;
	}
	*dst++ = '\0';

	/* require 1 ! & @, ban len >= BANLEN */
	if (!fl_ex || !fl_at || ((dst - ban_str) >= BANLEN)) return NULL;

	return ban_str;
}

/*
================
sha_string
----------------
convert a string into a SHA1 message digest
================
*/

char *sha_string (char *dst, const char *src)
{
#if defined(HAVE_SSL)
	byte md[SHA_DIGEST_LENGTH];
	int i;

	assert(dst);
	assert(src);

	/* create the message digest from the string */
	SHA1(src, strlen(src), md);

	/* convert digest into a hex string */
	for (i = 0; i < SHA_DIGEST_LENGTH; i++) sprintf(dst + i * 2, "%02x", (int) md[i]);
	dst[SHA_DIGEST_LENGTH * 2] = '\0';

	return dst;
#else
	assert(dst);
	assert(src);
	/* WARN: server passwords are NOT hashed! */
	strcpy(dst,src);
	return dst;
#endif /* HAVE_SSL */
}

/*
================
host_encrypt
----------------
convert the host into a SHA1 message digest
================
*/

char *host_encrypt (const char *host)
{
#ifdef HAVE_SSL
	byte md[SHA_DIGEST_LENGTH];
	char buf[6 + (SHA_DIGEST_LENGTH * 2)], *dst;
	int i;

	/* create the message digest from the host string */
	SHA1(host, strlen(host), md);

	/* set host prefix */
	strcpy(buf,"host-");
	dst = buf + 5;

	/* convert digest into a hex string */
	for (i = 0; i < 5; i++) sprintf(dst + i * 2, "%02x", (int) md[i]);

	return str_dup(buf);
#else
	/* No OpenSSL == No Host Encryption */
	char *str, *dst;
	int len;

	/* just extract the plain host name of the client */
	str = strchr(host,'.');
	if (!str) str = host + strlen(host);

	len = str - host;
	dst = sys_malloc(len + 1);
	memcpy(dst,host,len);
	dst[len] = '\0';

	return dst;
#endif /* HAVE_SSL */
}

/*
================
debug_memory
================
*/

void debug_memory (client_t *cl)
{
	u_int32_t n_cl = UserStats.N_users;
	u_int32_t n_ch = UserStats.N_chans;
	u_int32_t n_br = ip_branches();
	u_int32_t n_ip = ip_count(0,0);

	cl_send_fmt(cl, ":%s %.3d %s :Clients  - %u entries (%u) : %u bytes",
		Srv->name, (int) RPL_STATSDEBUG, cl->name,
		n_cl, sizeof(client_t), n_cl * sizeof(client_t));

	cl_send_fmt(cl, ":%s %.3d %s :Channels - %u entries (%u) : %u bytes",
		Srv->name, (int) RPL_STATSDEBUG, cl->name,
		n_ch, sizeof(channel_t), n_ch * sizeof(channel_t));

	cl_send_fmt(cl, ":%s %.3d %s :IP Cache - %u entries : %u branches (%u) : %u bytes",
		Srv->name, (int) RPL_STATSDEBUG, cl->name,
		n_ip, n_br, sizeof(ip_tree_t), n_br * sizeof(ip_tree_t));
}

/*
================
show_motd
================
*/

void show_motd (client_t *cl)
{
	FILE *file;
	char line[256];

	if (Options->motd && (file = fopen(Options->motd,"r")))
	{
		cl_send_numeric(cl, RPL_MOTDSTART, Srv->name);
		while ( fgets(line, sizeof(line), file) )
			cl_send_numeric(cl, RPL_MOTD, line);
		fclose(file);
	}
	cl_send_numeric(cl, RPL_ENDOFMOTD);
}

/*
================
chan_mode
================
*/

u_int32_t chan_mode (char ch)
{
	switch (ch)
	{
	/* channel modes */
#ifdef HAVE_SSL
	case 'x':
		return CM_SSLONLY;
#endif /* HAVE_SSL */
	case 's':
		return CM_SECRET;
	case 'm':
		return CM_MODERATED;
	case 'n':
		return CM_NOPRIVMSGS;
	case 't':
		return CM_TOPICLIMIT;
	case 'i':
		return CM_INVITEONLY;
	case 'k':
		return CM_KEY;
	case 'l':
		return CM_LIMIT;
	/* user channel modes */
	case 'o':
		return CF_OP;
	case 'v':
		return CF_VOICE;
	/* unknown */
	default:
		return 0;
	}
}

/*
================
user_mode
================
*/

u_int32_t user_mode (char ch)
{
	switch (ch)
	{
	/* user modes */
	case 'o':
		return FL_OPERATOR;
	case 'i':
		return FL_INVISIBLE;
	case 's':
		return FL_SNOTICE;
	case 'w':
		return FL_WALLOPS;
	/* enhanced modes */
	case 'k':
		return FL_SERVICE;
	/* unknown */
	default:
		return 0;
	}
}

