/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: ircd_plugins.c,v 1.11 2002/01/29 18:39:39 zarjazz Exp $
*/

#include <ircd_include.h>

#ifdef WINVER
	#pragma warning(disable : 4090)	// const
#else
	#include <dlfcn.h>
	#include <dirent.h>
	#include <sys/stat.h>
#endif

/* ================ */

/* global plugin ui */
cl_ui_t cl_ui;
sv_ui_t sv_ui;

/* plugin function typedef */
typedef cmd_entry_t * (*CL_PLUGIN_API) (cl_ui_t *ui);
typedef cmd_entry_t * (*SV_PLUGIN_API) (sv_ui_t *ui);

#ifdef WINVER
	typedef HINSTANCE PLUGIN_DLL;
#else
	typedef void *PLUGIN_DLL;
#endif

/* plugin handle table */
static int _plugin_table_size;
static PLUGIN_DLL *_plugin_table;

/*
================
ui helper function prototypes
================
*/

void _server_add (client_t *sv);
void _server_del (client_t *sv);

void _channel_add (channel_t *chan);
void _channel_del (channel_t *chan);

void _client_add (client_t *chan);
void _client_del (client_t *chan);

/*
================
load_plugin
================
*/

static int _cmd_count (cmd_entry_t *cstr)
{
	int count = 0;
	if (!cstr) return 0;
	while (cstr->cmd) cstr++, count++;
	return count;
}

static void load_plugin (const char *dll)
{
	CL_PLUGIN_API cl_api = NULL;
	SV_PLUGIN_API sv_api = NULL;
	PLUGIN_DLL hDLL;

	/* load the dynamic library */
#ifdef WINVER
	hDLL = LoadLibrary(dll);
	if (!hDLL) { sys_warn("unable to load plugin: %s\n%s\n",dll,sys_str_error()); return; }
#else
	hDLL = dlopen(dll, RTLD_NOW); /* in case there are some stupid people out there */
	if (!hDLL) { sys_warn("unable to load plugin: %s\n",dlerror()); return; }
#endif

	/* register plugin for cleanup */
	_plugin_table = sys_realloc(_plugin_table, sizeof(PLUGIN_DLL *) * (_plugin_table_size + 1));
	_plugin_table[_plugin_table_size++] = hDLL;

	/* client interface */
#ifdef WINVER
	cl_api = (CL_PLUGIN_API) GetProcAddress(hDLL, "init_client_interface");
#else
	cl_api = (CL_PLUGIN_API) dlsym(hDLL, "init_client_interface");
#endif
	if (cl_api)
	{
		int ncmd, nsrc;
		cmd_entry_t *src = cl_api(&cl_ui);

		/* add the commands to the list */
		ncmd = _cmd_count(CMD_ENTRY_LIST);
		nsrc = _cmd_count(src) + 1;

		CMD_ENTRY_LIST = sys_realloc(CMD_ENTRY_LIST, sizeof(cmd_entry_t) * (ncmd + nsrc));
		memcpy(CMD_ENTRY_LIST + ncmd, src, sizeof(cmd_entry_t) * nsrc);

		sys_dprintf(1,"loaded client interface: %s\n",dll);
	}

	/* server interface */
#ifdef WINVER
	sv_api = (SV_PLUGIN_API) GetProcAddress(hDLL, "init_server_interface");
#else
	sv_api = (SV_PLUGIN_API) dlsym(hDLL, "init_server_interface");
#endif
	if (sv_api)
	{
		MSG_ENTRY_LIST = sv_api(&sv_ui);
		sys_dprintf(1,"loaded server interface: %s\n",dll);
	}
}

/*
================
cleanup_plugins
================
*/

static void _cleanup_plugins (void)
{
	sys_dprintf(2,"Unloading plugin's\n");

	while (_plugin_table_size--)
	{
		PLUGIN_DLL hDLL;
		hDLL = _plugin_table[_plugin_table_size];
#ifdef WINVER
		FreeLibrary(hDLL);
#else
		dlclose(hDLL);
#endif
	}
	sys_free(_plugin_table);
	sys_free(CMD_ENTRY_LIST);

	sys_dprintf(2,"All modules unloaded\n");
}

static void cleanup_plugins (void)
{
	WRITE_LOCK(rwlock_commands);
	_cleanup_plugins();
	WRITE_UNLOCK(rwlock_commands);
	RWLOCK_DESTROY(rwlock_commands);
}

/*
================
scandir / alphasort
----------------
Sun don't implement these, like *how* lame?!
================
*/

#ifdef __sun__

int alphasort (const struct dirent **a, const struct dirent **b)
{
	return strcmp((*a)->d_name, (*b)->d_name);
}

int scandir (const char *dir, struct dirent ***namelist,
			int (*select)(const struct dirent *),
			int (*compar)(const struct dirent **, const struct dirent **))
{
	DIR *d;
	struct dirent *entry;
	size_t entrysize;
	int i = 0;

	if ((d=opendir(dir)) == NULL) return -1;

	*namelist = NULL;
	while ((entry=readdir(d)) != NULL)
	{
		if (select == NULL || (select != NULL && (*select)(entry)))
		{
			*namelist = (struct dirent **)realloc((void *)(*namelist),
						(size_t)((i+1)*sizeof(struct dirent *)));
			if (*namelist == NULL) return -1;
			entrysize = sizeof(struct dirent) - sizeof(entry->d_name) + strlen(entry->d_name) + 1;
			(*namelist)[i] = (struct dirent *) malloc(entrysize);
			if ((*namelist)[i] == NULL) return -1;
			memcpy((*namelist)[i], entry, entrysize);
			i++;
		}
	}
	if (closedir(d)) return -1;
	if (i == 0) return -1;
	if (compar != NULL) qsort((void *)(*namelist), (size_t) i, sizeof(struct dirent *), compar);

	return i;
}

#endif /*__sun__*/

/*
================
load_plugins
================
*/

void load_plugins (bool lock, bool reload)
{
	char *dir;
	char path[1024];
	block_arg_t *range = NULL;

	if (lock) WRITE_LOCK(rwlock_commands);
	if (reload) _cleanup_plugins();

	/* initialize structure */
	_plugin_table_size = 0;
	_plugin_table = NULL;

	CMD_ENTRY_LIST = NULL;
	MSG_ENTRY_LIST = NULL;

	/* go through plugin directories, loading all libraries */
	while ( range = arguments_by_name(Config, range, "PluginDir") )
	{
#ifdef WINVER
		WIN32_FIND_DATA fileData;
		HANDLE findHandle;

		dir = range->arg_value;
	#ifdef _DEBUG
		snprintf(path,sizeof(path),"%s\\*_debug.dll",dir);
	#else
		snprintf(path,sizeof(path),"%s\\*.dll",dir);
	#endif
		findHandle = FindFirstFile(path, &fileData);
		if (findHandle == INVALID_HANDLE_VALUE)
		{
			sys_warn("Unable to find plugins: %s\n",dir);
			continue;
		}

		do
		{
			if (fileData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) continue;
			if (fileData.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM) continue;
			if (fileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) continue;

			snprintf(path,sizeof(path),"%s\\%s",dir,fileData.cFileName);
			load_plugin(path);
		} while ( FindNextFile(findHandle, &fileData) );
#else
		int i,n;
		char *fname;
		struct stat st;
		struct dirent **namelist;

		dir = range->arg_value;
		n = scandir(dir, &namelist, 0, alphasort);
		if (n < 0)
		{
			sys_warn("Unable to find plugins: %s\n",dir);
			continue;
		}

		for (i = 0; i < n; i++)
		{
			fname = namelist[i]->d_name;

			if ( !str_match(fname, "*.so", false) ) continue;

			snprintf(path,sizeof(path),"%s/%s",dir,fname);
			if (stat(path,&st) < 0) continue;
			if (S_ISDIR(st.st_mode)) continue;

			load_plugin(path);
		}

		for (i = 0; i < n; i++) free(namelist[i]);
		free(namelist);
#endif /* WINVER */
	}
	if (lock) WRITE_UNLOCK(rwlock_commands);
}

/*
================
init_plugins
================
*/

void init_plugins (void)
{
	RWLOCK_INIT(rwlock_commands);

	/* initialize client interface */
	memset(&cl_ui,0,sizeof(cl_ui));
	memset(&sv_ui,0,sizeof(sv_ui));
	memset(&UserStats,0,sizeof(UserStats));
	/* public variables */
	cl_ui.VERSION			= VERSION;
	cl_ui.Now				= &Now;
	cl_ui.Srv				= Srv;
	cl_ui.Options			= Options;
	cl_ui.Config			= &Config;
	cl_ui.UserStats			= &UserStats;
	/* public network functions */
	cl_ui.cl_send_raw		= cl_send_raw;
	cl_ui.cl_send_fmt		= cl_send_fmt;
	cl_ui.cl_send_numeric	= cl_send_numeric;
	cl_ui.cl_send_names		= cl_send_names;
	cl_ui.cl_send_njoin		= cl_send_njoin;
	cl_ui.cl_send_npart		= cl_send_npart;
	cl_ui.chan_send_fmt		= chan_send_fmt;
	cl_ui.chan_send_numeric	= chan_send_numeric;
	cl_ui.all_send_fmt		= all_send_fmt;
	/* public config functions */
	cl_ui.config_string		= config_string;
	/* public client functions */
	cl_ui.client_exit		= client_exit;
	cl_ui.client_userhost	= client_userhost;
	cl_ui.client_userip		= client_userip;
	cl_ui.client_register	= client_register;
	cl_ui.client_find		= client_find;
	cl_ui.client_find_invite= client_find_invite;
	cl_ui.client_find_chan	= client_find_chan;
	cl_ui.client_set_nick	= client_set_nick;
	cl_ui.client_idle_time	= client_idle_time;
	cl_ui.client_cookie		= client_cookie;
	cl_ui.client_pong		= client_pong;
	cl_ui.client_mode_str	= client_mode_str;
	cl_ui.client_parse_mode_str = client_parse_mode_str;
	/* public channel functions */
	cl_ui.channel_exit		= channel_exit;
	cl_ui.channel_find		= channel_find;
	cl_ui.channel_find_nick = channel_find_nick;
	cl_ui.channel_mode_str	= channel_mode_str;
	cl_ui.channel_parse_mode_str = channel_parse_mode_str;
	cl_ui.channel_join		= channel_join;
	cl_ui.channel_list		= channel_list;
	/* public misc stuff */
	cl_ui.invite_remove		= invite_remove;
	cl_ui.ban_remove		= ban_remove;
	cl_ui.ban_add			= ban_add;
	cl_ui.ircd_terminate	= ircd_terminate;
	cl_ui.ircd_rehash		= ircd_rehash;
	cl_ui.valid_nick_char	= valid_nick_char;
	cl_ui.valid_chan_char	= valid_chan_char;
	cl_ui.sanify_ban		= sanify_ban;
	cl_ui.debug_memory		= debug_memory;
	cl_ui.show_motd			= show_motd;
	cl_ui.chan_mode			= chan_mode;
	cl_ui.user_mode			= user_mode;
	cl_ui.server_notice		= server_notice;
	cl_ui.server_match		= server_match;

	/* initialize server interface */
	sv_ui.Now				= &Now;
	sv_ui.Srv				= Srv;
	/* public functions */
	sv_ui.server_exit		= client_exit;
	sv_ui.server_copy		= server_copy;
	sv_ui.server_register	= server_register;
	sv_ui.server_password	= server_password;
	sv_ui.server_create		= server_create;
	sv_ui.server_malloc		= server_malloc;
	sv_ui.server_find		= client_find;
	sv_ui.server_find_id	= server_find_id;
	sv_ui.server_notice		= server_notice;
	sv_ui.server_ping		= client_ping;
	sv_ui.server_pong		= client_pong;
	/* public network functions */
	sv_ui.sv_send_raw		= cl_send_raw;
	sv_ui.sv_send_fmt		= cl_send_fmt;
	sv_ui.sv_send_notice	= server_send_notice;
	sv_ui.sv_send_numeric	= cl_send_numeric;
	sv_ui.chan_send_fmt		= chan_send_fmt;
	/* public client functions */
	sv_ui.client_exit		= client_exit;
	sv_ui.client_userhost	= client_userhost;
	sv_ui.client_userip		= client_userip;
	sv_ui.client_find		= client_find;
	sv_ui.client_create_remote = client_create_remote;
	sv_ui.client_find_chan	= client_find_chan;
	sv_ui.client_parse_mode_str = client_parse_mode_str;
	/* public channel functions */
	sv_ui.channel_create	= channel_create;
	sv_ui.channel_find		= channel_find;
	sv_ui.channel_find_nick = channel_find_nick;
	sv_ui.channel_parse_mode_str = channel_parse_mode_str;
	sv_ui.channel_join		= channel_join;
	sv_ui.channel_exit		= channel_exit;
	/* sync */
	sv_ui.sync_start		= sync_start;
	/* public misc stuff */
	sv_ui.server_link_add	= server_link_add;
	sv_ui.server_link_del	= server_link_del;
	sv_ui.valid_nick_char	= valid_nick_char;
	sv_ui.valid_chan_char	= valid_chan_char;

	/* loading all plugins */
	load_plugins(true,false);

	/* register for cleanup */
	sys_cleanup_register(cleanup_plugins);
}
