/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: ircd_servers.c,v 1.15 2002/01/27 22:34:36 zarjazz Exp $
*/

#define __USE_SERVER_STRUCT
#include <ircd_include.h>

/* ================ */

static client_t *_ids[NET_MAX_SERVERS];

rwlock_t rwlock_servers;

options_t *Options;
client_t *Srv;

void _client_add (client_t *chan);
void _client_del (client_t *chan);

/*
================
get_server_ids
================
*/

client_t **get_server_ids (void) { return _ids; }

/*
================
_server_add
================
*/

void _server_add (client_t *sv)
{
	_client_add(sv);

	if (sys_debug_get())
	{
		char buf[128];
		net_str_sockaddr(buf,&sv->sa);
		sys_dprintf(2,"server (%u) add: %s\n", sv->id, buf);
	}

	_ids[sv->node.id] = sv;
	UserStats.N_servs++;
}

/*
================
_server_del
================
*/

void _server_del (client_t *sv)
{
	if (sys_debug_get())
	{
		char buf[128];
		net_str_sockaddr(buf,&sv->sa);
		sys_dprintf(2,"server (%u) del: %s\n", sv->id, buf);
	}

	UserStats.N_servs--;
	sys_free(sv->node.neighbours);
	_client_del(sv);
	server_close(sv);
}

/*
================
_clean_array
================
*/

static void _clean_array (array_t *array)
{
	if (!array) return;

	while (array->nitems)
	{
		char *item;

		item = *((char **) array->list);
		array_del_item(array,&item,sizeof(char *));
		sys_free(item);
	}
}

/*
================
cleanup_local
================
*/

static void cleanup_local (void)
{
	_clean_array(&Options->rsv_nicks);
	_clean_array(&Options->rsv_chans);
}

/*
================
server_malloc
================
*/

client_t *server_malloc (u_int16_t id)
{
	client_t *sv = sys_malloc(sizeof(*sv));
	sv->tree_key = sv->name;
	sv->local.server = sv;
	sv->node.neighbours = sys_malloc(sizeof(u_int16_t) * NET_MAX_SERVERS);
	sv->node.id = id;
	bit_set(sv,FL_SERVER);
	return sv;
}

/*
================
local_server_id
----------------
return the id of the clients local server
================
*/

u_int16_t local_server_id (client_t *cl) { return cl->local.server->node.id; }

/*
================
server_links
================
*/

void server_link_add (client_t *sv, u_int16_t id) { link_add(&sv->node,id); }
void server_link_del (client_t *sv, u_int16_t id) { link_del(&sv->node,id); }

/*
================
server_route
================
*/

client_t *server_route (client_t *sv)
{
	node_t *node = node_find_route(sv->node.id);
	return _ids[node->id];
}

/*
================
init_local
----------------
initialize the local server
================
*/

void init_local (void)
{
	block_t *block = NULL;
	block_t *root = Config;
	block_arg_t *blk_arg;
	char *str;

	Options = sys_malloc(sizeof(*Options));

	/* create the local server structure, id */
	str = ircd_config_string(root, NULL, NULL, "ServerID");
	Srv = server_malloc((u_int16_t) (atoi(str) % NET_MAX_SERVERS));
	sys_dprintf(1, "ServerID: %hu\n", Srv->node.id);

	/* server name */
	str_copy(Srv->name, ircd_config_string(root, NULL, NULL, "ServerName"), sizeof(Srv->name));
	sys_dprintf(1, "ServerName: %s\n", Srv->name);

	/* server description */
	str_copy(Srv->info, ircd_config_string(root, NULL, NULL, "Description"), sizeof(Srv->info));

	/* set the dynamic variables */
	server_options(root,Options);

	/* extract reserved nick list */
	array_init(&Options->rsv_nicks,4);

	block = NULL;
	blk_arg = NULL;
	while ( block = blocks_by_name(root,block,"Reserve") )
		while ( blk_arg = arguments_by_name(block,blk_arg,"Nicks") )
			argument_list(blk_arg,&Options->rsv_nicks);

	/* server id array */
	memset(_ids,0,sizeof(_ids));
	_ids[0] = Srv;

	/* SSL */
#ifdef HAVE_SSL
	str = ircd_config_string(root, NULL, NULL, "SSL");
	if (str && !stricmp(str,"true")) bit_set(Srv,FL_SSL);
#endif /* HAVE_SSL */

	/* create server */
	server_create(Srv, true);

	/* register for cleanup */
	sys_cleanup_register(cleanup_local);

	/* initialize the nodes */
	init_nodes(&Srv->node, NET_MAX_SERVERS);
}

/*
================
server_options
----------------
set all the server config dynamic variables
================
*/

void server_options (block_t *root, options_t *opt)
{
	char *str;

	/* motd filename */
	opt->motd = ircd_config_string(root, NULL, NULL, "MOTD");

	/* network */
	opt->network = ircd_config_string(root, NULL, NULL, "Network");
	sys_dprintf(1, "Network: %s\n", opt->network);

	/* client host settings */
	str = ircd_config_string(root, NULL, NULL, "ClientHosts");
	if ( !stricmp(str,"encrypt") )
	{
		opt->encrypt_hosts	= true;
		opt->hide_domain	= false;
		opt->hide_ip		= true;
	}
	else if ( !stricmp(str,"hidden") )
	{
		opt->encrypt_hosts	= false;
		opt->hide_domain	= true;
		opt->hide_ip		= true;
	}
	else /* if ( !stricmp(str,"visible") ) */
	{
		opt->encrypt_hosts	= false;
		opt->hide_domain	= false;
		opt->hide_ip		= false;
	}
	sys_dprintf(1, "ClientHosts:\n");
	sys_dprintf(1, "  Encrypt Hosts = %s\n", opt->encrypt_hosts ? "true" : "false");
	sys_dprintf(1, "  Hide Domain   = %s\n", opt->hide_domain ? "true" : "false");
	sys_dprintf(1, "  Hide IP       = %s\n", opt->hide_ip ? "true" : "false");

	/* are normal users allowed to set +w? */
	str = ircd_config_string(root, NULL, NULL, "LuserWallops");
	opt->luser_wallops = (str && !stricmp(str,"true")) ? true : false;
	sys_dprintf(1, "LuserWallops: %s\n", opt->luser_wallops ? "true" : "false");

	/* send a notice about user join & quits? */
	str = ircd_config_string(root, NULL, NULL, "LuserConnect");
	opt->luser_connect = (str && !stricmp(str,"true")) ? true : false;
	sys_dprintf(1, "LuserConnect: %s\n", opt->luser_connect ? "true" : "false");

	/* max users */
	str = ircd_config_string(root, NULL, NULL, "MaxUsers");
	opt->max_users = str ? atoi(str) : 0;
	sys_dprintf(1, "MaxUsers: %u\n", opt->max_users);

	/* max clones */
	str = ircd_config_string(root, NULL, NULL, "MaxClones");
	opt->max_clones = str ? atoi(str) : 0;
	sys_dprintf(1, "MaxClones: %u\n", opt->max_clones);

	/* max channels */
	str = ircd_config_string(root, NULL, NULL, "MaxChannels");
	opt->max_channels = str ? atoi(str) : 12;
	sys_dprintf(1, "MaxChannels: %u\n", opt->max_channels);

	/* max bans */
	str = ircd_config_string(root, NULL, NULL, "MaxBans");
	opt->max_bans = str ? atoi(str) : 30;
	sys_dprintf(1, "MaxBans: %u\n", opt->max_bans);

	/* ping freq */
	str = ircd_config_string(root, NULL, NULL, "PingFreq");
	opt->ping_freq = str ? atoi(str) : 90;
	sys_dprintf(1, "PingFreq: %u\n", opt->ping_freq);
}

/*
================
server_create_socket
================
*/

void server_create_socket (client_t *sv)
{
	struct sockaddr_in sa;
	char msg[BUFSIZE];
	int rc;

	net_sockaddr(&sa,NULL,0);

	rc = net_sock_tcp(&sa);
	if (rc < 0)
	{
		net_str_sockaddr(msg,&sa);
		sys_error("creating server socket: %s", msg);
	}
	else
	{
		sv->sock = rc;
		net_sock_buffer (sv->sock, SERVER_BUFFER);
	}
}

/*
================
server_notice
================
*/

void server_notice (const char *msg, const char *arg)
{
	all_send_fmt(FL_SNOTICE, ":%s NOTICE %s :%s: %s", Srv->name, Srv->name, msg, arg ? arg : "");
}

/*
================
server_send_notice
================
*/

void server_send_notice (client_t *sv, const char *msg, const char *arg)
{
	cl_send_fmt(sv, "NOTICE %s :%s: %s", Srv->name, msg, arg ? arg : "");
}

/*
================
server_find_id
----------------
MUST call with a client lock
================
*/

client_t *server_find_id (u_int16_t id) { return id ? _ids[id] : NULL; }

/*
================
server_address
================
*/

void server_address (struct sockaddr_in *sa, const char *address)
{
	u_int16_t port = 4400;
	const char *str_addr;
	char *str_port;

	str_port = strchr(address,':');
	if (str_port) { *str_port = '\0'; port = atoi(str_port + 1); }

	str_addr = strcmp(address, "*") ? address : NULL;
	net_sockaddr(sa,str_addr,port);

	if (str_port) *str_port = ':';
}

/*
================
server_close
================
*/

void server_close (client_t *sv)
{
	bit_set(sv, FL_ZOMBIE);

#ifdef HAVE_SSL
	if (sv->ssl)
	{
		SSL_shutdown(sv->ssl);
		SSL_free(sv->ssl);
		sv->ssl = NULL;
	}
#endif /* HAVE_SSL */
	net_sock_close(sv->sock);
	sv->sock = 0;

	sv->tv_connect = Now.tv_sec + 60 * 3;

	if (bit_tst(sv,FL_CONNECTED)) { _ids[sv->node.id] = NULL; sv->node.id = 0; }
}

/*
================
server_password
================
*/

char *server_password (client_t *sv)
{
	block_t *block = NULL;
	block_t *root = Config;

	/* step through Server blocks */
	while ( block = blocks_by_name(root,block,"Server") )
	{
		/* match server name */
		if ( stricmp(sv->name, block->block_value) ) continue;

		/* match password */
		return argument_by_name(block,"Password");
	}

	return NULL;
}

/*
================
server_copy
----------------
move the details from the temporary accepted server structure
================
*/

void server_copy (client_t *dst, client_t *src, u_int16_t id)
{
	if (!dst->node.id)
	{
			_ids[id] = dst;
			dst->node.id = id;
			UserStats.N_servs++;
	}
	bit_clr(src,FL_CONNECTED);
	dst->domain = src->domain;
	dst->tv_connect = src->tv_connect;
	dst->sa = src->sa;
	dst->sock = src->sock;
	src->sock = 0;
#ifdef HAVE_SSL
	dst->ssl = src->ssl;
	src->ssl = NULL;
#endif /* HAVE_SSL */
	bit_clr(dst,FL_ZOMBIE);
}

/*
================
server_connect
----------------
link local server to another
================
*/

int server_connect (client_t *sv)
{
	int rc;
#ifdef HAVE_SSL
	SSL *ssl = NULL;
#endif /* HAVE_SSL */

	rc = connect(sv->sock, (struct sockaddr *) &sv->sa, sizeof(sv->sa));
	if (rc < 0)
	{
		server_notice("Connect Failed",sv->name);
		server_close(sv);
		return -1;
	}

	/* if SSL server, do the handshake */
#ifdef HAVE_SSL
	if (bit_tst(sv,FL_SSL))
	{
		ssl = SSL_new(Ctx);
		if (!ssl) goto ssl_err;

		rc = SSL_set_fd(ssl, sv->sock);
		if (rc != 1) goto ssl_err;

		rc = SSL_connect(ssl);
		if (rc != 1)
		{
			char msg[BUFSIZE];
			unsigned long err;

			sys_warn("SSL server connect failed: %s %d", sv->name, rc);
			while (err = ERR_get_error()) sys_warn("  %s", ERR_error_string(err, msg));

			goto ssl_err;
		}
		sv->ssl = ssl;
	}
#endif /* HAVE_SSL */

	/* server connected sucessfully */
	bit_set(sv,FL_LOCAL | FL_CONNECTED);
	sv->tv_connect = Now.tv_sec;
	server_notice("Server Connected", sv->name);

	/* send PASS/SERVER packets */
	cl_send_fmt(sv, "PASS %s %s %s %s", server_password(sv), "0300", "SIRCD|", "");
	cl_send_fmt(sv, "SERVER %s 0 %d :sircd %s", Srv->name, Srv->node.id, VERSION);

	return 0;

#ifdef HAVE_SSL
ssl_err:
	if (ssl) { SSL_shutdown(ssl); SSL_free(ssl); }
	server_notice("SSL Connect Failed", sv->name);
	server_close(sv);
	return -1;
#endif /* HAVE_SSL */
}

/*
================
server_register
================
*/

void server_register (client_t *sv)
{
	bit_set(sv,FL_REGISTERED);

	/* add link to the mesh */
	server_link_add(Srv, sv->node.id);
}

/*
================
server_match
================
*/

client_t *server_match (const char *str)
{
	client_t *sv;
	int i;

	for (i = 1; i < NET_MAX_SERVERS; i++)
	{
		sv = _ids[i];
		if (!sv) continue;
		if ( !bit_tst(sv,FL_REGISTERED) ) continue;
		if ( str_match(sv->name, str, false) ) return sv;
	}

	return NULL;
}
