/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: ircd_sync.c,v 1.8 2002/01/27 22:34:36 zarjazz Exp $
*/

#define __USE_SERVER_STRUCT
#include <ircd_include.h>

/* Prototypes */
tree_t *	get_channels	(void);
tree_t *	get_clients		(void);
client_t **	get_server_ids	(void);

/*
================
sync_nick
================
*/

void send_nick (client_t *sv, client_t *cl)
{
	char userhost[USERHOSTLEN];	/* nick!user@host of client */
	char umode[32];

	/* no need to send details about a client thats local to the destination server */
	if (cl->local.server == sv) return;

	client_userhost(cl,userhost,true); /* hide details from remote servers */
	client_mode_str(umode,'+',cl->flags);

	cl_send_fmt(sv, "USER %s %hu %s :%s", userhost, cl->local.server->node.id, umode, cl->info);
}

/*
================
sync_clients
================
*/

static void __sync_clients (client_t *sv, byte *buf, tree_t *tree)
{
	client_t *cl = (client_t *) tree;

	if (!cl) return;

	if (!bit_tst(cl,FL_ZOMBIE) && is_user(cl))
	{
		if (bit_tst(cl,FL_SYNCING))
			bit_clr(cl,FL_SYNCING);
		else
			send_nick(sv,cl);
	}
	__sync_clients(sv, buf, tree->left);
	__sync_clients(sv, buf, tree->right);
}

void sync_clients (client_t *sv, byte *buf)
{
	/* sync all clients */
	__sync_clients(sv, buf, get_clients());
}

/*
================
sync_channels
================
*/

void __sync_channels (client_t *sv, byte *buf, tree_t *tree)
{
	channel_t *chan = (channel_t *) tree;

	if (!chan) return;

	if (!bit_tst(chan,FL_ZOMBIE))
	{
		char umode[128];
		index_t *index;
		int len;

		/* send all unsync'd channel members */
		for (index = (index_t *) chan->members.item_start; index; index = index->next)
		{
			client_t *cl = (client_t *) index->item;

			/* only send valid clients */
			if (bit_tst(cl,FL_ZOMBIE)) continue;

			/* if we haven't sent this clients details yet, do so */
			if (!bit_tst(cl,FL_SYNCING))
			{
				send_nick(sv,cl);
				bit_set(cl,FL_SYNCING);
			}
		}

		/* send channel details */
		channel_mode_str(umode, '+', chan->flags, chan->key, chan->limit);
		cl_send_fmt(sv,":%hu SYNC %s %s", Srv->node.id, chan->name, umode);

		len = sprintf(buf,":%hu SYNC NAMES %s :", Srv->node.id, chan->name);
		do
		{
			int n;
			n = channel_names(buf + len, BUFSIZE - len, chan, &index, true);
			cl_send_raw(sv,buf,len + n);
		} while (index);
	}

	__sync_channels(sv, buf, tree->left);
	__sync_channels(sv, buf, tree->right);
}

static void sync_channels (client_t *sv, byte *buf)
{
	__sync_channels(sv, buf, get_channels());
}

/*
================
sync_server
----------------
Pass on client / channel information to a new server,
ie one not linked anywhere to the network
================
*/

static int sync_server (client_t *dst, bool request)
{
	client_t **ids = get_server_ids();
	char buf[BUFSIZE];
	int i;

	sys_dprintf(1, "Syncing: %s (%hu)\n", dst->name, dst->node.id);

	/* request remote sync data? */
	if (request) cl_send_fmt(dst,":%hu SYNC REQUEST\r\n", Srv->node.id);

	/* send the server list first */
	for (i = 1; i < NET_MAX_SERVERS; i++)
	{
		client_t *sv;

		sv = ids[i];
		/* skip null & destination server */
		if (!sv || i == dst->node.id) continue;

		if (sv == Srv || bit_tst(sv,FL_REGISTERED))
			cl_send_fmt(dst, ":%hu SERVER %s 0 %hu :sircd %s", Srv->node.id, sv->name, sv->node.id, VERSION);
	}

	/* send server link state */
	for (i = 1; i < NET_MAX_SERVERS; i++)
	{
		client_t *sv;

		sv = ids[i];
		/* skip null & destination server */
		if (!sv || i == dst->node.id) continue;

		if (sv == Srv || bit_tst(sv,FL_REGISTERED))
		{
			if (sv->node.num_neighbours)
			{
				link_state_string(buf,&sv->node);
				cl_send_fmt(dst, ":%hu LINKS%s", Srv->node.id, buf);
			}
		}
	}

	/* sync channels */
	sync_channels(dst,buf);

	/* sync remaining clients */
	sync_clients(dst,buf);

	/* sync complete */
	cl_send_fmt(dst,":%hu SYNC DONE\r\n", Srv->node.id);

	return 0;
}

/*
================
callback_sync
================
*/

void callback_sync (void *arg1, void *arg2)
{
	u_int32_t id = (u_int32_t) arg1;
	bool request = (bool) arg2;
	client_t *sv;

	WRITE_LOCK(rwlock_clients);
	WRITE_LOCK(rwlock_channels);
	sv = client_validate(id);
	if (sv) sync_server(sv,request);
	WRITE_UNLOCK(rwlock_channels);
	WRITE_UNLOCK(rwlock_clients);
}

/*
================
sync_start
================
*/

void sync_start (client_t *sv, char *prefix, bool request)
{
	/* set node id from request prefix if required */
	if (!sv->node.id && prefix)
	{
		u_int16_t id = atoi(prefix);
		client_t **ids = get_server_ids();

		sv->node.id = id;
		ids[id] = sv;
	}

	task_create(callback_sync, (void *) sv->id, (void *) request);
}
