/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: ircd_threads.c,v 1.10 2002/01/15 23:04:42 zarjazz Exp $
*/

#ifdef WINVER
  #pragma warning(disable : 4761)	// conversion
#endif

#include <ircd_include.h>

/* ================ */

static llist_t _listeners;
static llist_t _workers;
static llist_t _tasks;

static mutex_t mutex_workers;
static mutex_t mutex_tasks;

static cond_t cond_tasks;

listener_t *Remote;

/* ircd_servers.c */
void _server_add (client_t *sv);
void _server_del (client_t *sv);

/*
================
cleanup_threads
================
*/

static void cleanup_threads (void)
{
	listener_t *listener;
	worker_t *worker;
	task_t *task;
	void *next;

	/* ================ */

	for (worker = (worker_t *) _workers.item_start; worker; worker = worker->next)
	{
		worker->die = true;
		task_create(NULL,NULL,NULL);
	}

	for (worker = (worker_t *) _workers.item_start; worker; worker = next)
	{
		next = worker->next;
		THREAD_JOIN(worker->thread);
		ll_del_item(&_workers,worker);
		sys_free(worker);
	}

	MUTEX_DESTROY(mutex_workers);

	/* ================ */

	for (listener = (listener_t *) _listeners.item_start; listener; listener = listener->next) listener->die = true;

	for (listener = (listener_t *) _listeners.item_start; listener; listener = next)
	{
		next = listener->next;
		ll_del_item(&_listeners,listener);
		if (listener->sock)
		{
			THREAD_JOIN(listener->thread);
			net_sock_close(listener->sock);
		}
		sys_free(listener);
	}

	/* ================ */

	MUTEX_LOCK(mutex_tasks);
	for (task = (task_t *) _tasks.item_start; task; task = next)
	{
		next = task->next;
		ll_del_item(&_tasks,task);
		sys_free(task);
	}
	MUTEX_UNLOCK(mutex_tasks);
	MUTEX_DESTROY(mutex_tasks);

	COND_DESTROY(cond_tasks);

	/* ================ */

	sys_dprintf(2,"All child threads destroyed\n");
}

/*
================
init_threads
================
*/

void init_threads (void)
{
	memset(&_listeners,0,sizeof(_listeners));
	memset(&_workers,0,sizeof(_workers));
	memset(&_tasks,0,sizeof(_tasks));

	MUTEX_INIT(mutex_workers);
	MUTEX_INIT(mutex_tasks);

	COND_INIT(cond_tasks);

	// register for cleanup
	sys_cleanup_register(cleanup_threads);
}

/*
================
thread_sigmask
----------------
setup thread signal stuff
================
*/

static void thread_sigmask (void)
{
#ifndef WINVER
	sigset_t set;

	sigemptyset(&set);
//	sigaddset(&set,SIGSEGV);
	sigaddset(&set,SIGTERM);
	sigaddset(&set,SIGHUP);
//	sigaddset(&set,SIGFPE);
//	sigaddset(&set,SIGILL);
	sigaddset(&set,SIGINT);

	pthread_sigmask(SIG_BLOCK,&set,NULL);
#endif
}

/*
================
task_submit
================
*/

void task_submit (task_t *task)
{
	MUTEX_LOCK(mutex_tasks);
	ll_add_end(&_tasks,task);
	MUTEX_UNLOCK(mutex_tasks);
	/* signal workers there is a new task */
	COND_SIGNAL(cond_tasks);
}

/*
================
task_create
================
*/

void task_create (void (* callback) (void *, void *), void *arg1, void *arg2)
{
	task_t *task = sys_malloc(sizeof(*task));

	task->callback = callback;
	task->arg1 = arg1;
	task->arg2 = arg2;
	task_submit(task);
}

/*
================
worker_thread
================
*/

THREAD_FUNC(worker_thread)
{
	worker_t *worker = arg;
	task_t *task;

	thread_sigmask();

	sys_dprintf(2,"Worker thread created\n");

	MUTEX_LOCK(mutex_tasks);
	while (!terminate && !worker->die)
	{
		if (!_tasks.nitems) COND_WAIT(cond_tasks,mutex_tasks);
		/* Note: we have the lock now but we may of missed the signaled data so double check */
		task = (task_t *) _tasks.item_start;
		if (!task) continue;

		ll_del_item(&_tasks,task);
		MUTEX_UNLOCK(mutex_tasks);
		/* execute callback function */
		if (task->callback) task->callback(task->arg1,task->arg2);
		sys_free(task);
		MUTEX_LOCK(mutex_tasks);
	}
	MUTEX_UNLOCK(mutex_tasks);

	sys_dprintf(2,"Worker thread killed\n");

	return 0;
}

/*
================
create_workers
================
*/

void create_workers (int count)
{
	int i;
	worker_t *worker;

	/* Lock just in case we die during startup */
	MUTEX_LOCK(mutex_workers);
	for (i = 0; i < count; i++)
	{
		worker = sys_malloc(sizeof(*worker));
		ll_add_end(&_workers,worker);
		THREAD_CREATE(worker->thread,worker_thread,(void *) worker);
	}
	MUTEX_UNLOCK(mutex_workers);
}

/*
================
listener_thread
================
*/

THREAD_FUNC(listener_thread)
{
	listener_t *listener = arg;

	thread_sigmask();

	sys_dprintf(2,"Listen thread created\n");
	while (!terminate && !listener->die) net_loop_listener(listener);
	sys_dprintf(2,"Listen thread killed\n");

	return 0;
}

/*
================
create_listener
================
*/

static void create_listener (listener_t *listener)
{
	char host[512];

	assert(listener);

	ll_add_end(&_listeners,listener);
	THREAD_CREATE(listener->thread,listener_thread,(void *) listener);
#ifdef HAVE_SSL
	sys_dprintf(2,"Listen: %s %s\n", net_str_sockaddr(host,&listener->sa), listener->use_ssl ? "(SSL)" : "");
#else
	sys_dprintf(2,"Listen: %s\n", net_str_sockaddr(host,&listener->sa));
#endif /* HAVE_SSL */
}

/*
================
create_listeners
================
*/

void create_listeners (block_t *root)
{
	listener_t *listener;
	int rc, nlisten = 0;

	block_t *block = NULL;

	/* create a dummy listener for remote clients */
	Remote = sys_malloc(sizeof(*listener));
	ll_add_end(&_listeners,Remote);

	/* create the listener threads */
	while ( block = blocks_by_name(root,block,"Listen") )
	{
		bool listen_server;
		u_int32_t user_max;
#ifdef HAVE_SSL
		bool use_ssl;
#endif /* HAVE_SSL */
		block_arg_t *blk_arg;
		char *host, *blk_val;

		if (strcmp("*",block->block_value))
			host = block->block_value;
		else
			host = NULL;

		/* is this listener for clients or servers? */
		blk_val = argument_by_name(block,"Type");
		listen_server = stricmp(blk_val,"Server") ? false : true;

		/* does this listen group require SSL? */
#ifdef HAVE_SSL
		blk_val = argument_by_name(block,"SSL");
		use_ssl = (blk_val && !stricmp(blk_val,"true")) ? true : false;
#endif /* HAVE_SSL */

		/* does this listen group require a max user limit? */
		blk_val = argument_by_name(block,"MaxUsers");
		user_max = blk_val ? atoi(blk_val) : 0;

		blk_arg = NULL;
		while ( blk_arg = arguments_by_name(block,blk_arg,"Port") )
		{
			u_int16_t port;
			array_t array;
			int32_t i;

			array_init(&array,8);

			/* allocate over the port ranges */
			argument_range(blk_arg,&array);

			for (i = 0; i < array.nitems; i++)
			{
				char buf[256];

				port = ((int *) array.list)[i];

				listener = sys_malloc(sizeof(*listener));

				net_sockaddr(&listener->sa,host,port);
				rc = net_sock_tcp(&listener->sa);
				if (rc < 0) sys_error("creating listener socket: %s",net_str_sockaddr(buf,&listener->sa));

				if (listen_server)
					net_sock_buffer(rc, SERVER_BUFFER);
#ifdef HAVE_SSL
				else if (use_ssl)
					net_sock_buffer(rc, CLIENT_BUFFER_SSL);
				else
#endif /* HAVE_SSL */
					net_sock_buffer(rc, CLIENT_BUFFER);

				net_sock_noblock(rc);
				listen(rc,SOMAXCONN);

				listener->sock = rc;
				listener->user_max = user_max;
#ifdef HAVE_SSL
				listener->use_ssl = use_ssl;
#endif /* HAVE_SSL */
				if (listen_server) bit_set(listener,FL_SERVER);

				create_listener(listener);
				nlisten++;
			}

			sys_free(array.list);
		}
	}
}

/*
================
server_thread
================
*/

THREAD_FUNC(server_thread)
{
	client_t *sv = arg;

	thread_sigmask();

	net_loop_server(sv);

	sys_dprintf(2,"Server thread killed\n");
	return 0;
}

/*
================
server_create
================
*/

void server_create (client_t *sv, bool lock)
{
	assert(sv);

	if (lock) WRITE_LOCK(rwlock_clients);
	sv->tree_key = sv->name;
	_server_add(sv);
	/* sendq */
	sendq_init(&sv->sendq);
	if (lock) WRITE_UNLOCK(rwlock_clients);
}

/*
================
create_servers
----------------
create all the leaf server threads
================
*/

void create_servers (block_t *root)
{
	block_t *block = NULL;
	client_t *sv;

	while ( block = blocks_by_name(root,block,"Server") )
	{
		char *blk_val;

		sv = server_malloc(0);

		/* specify this server is a local link */
		bit_set(sv,FL_LOCAL);

		/* leaf server name */
		str_copy(sv->name, block->block_value, sizeof(sv->name));

		/* leaf server address */
		blk_val = argument_by_name(block, "ServerAddress");
		server_address(&sv->sa,blk_val);

		/* SSL */
#ifdef HAVE_SSL
		blk_val = argument_by_name(block, "SSL");
		if (blk_val && !stricmp(blk_val,"true")) bit_set(sv,FL_SSL);
#endif /* HAVE_SSL */

		server_create(sv, true);

		/* create the server thread */
		THREAD_CREATE(sv->local.thread,server_thread,sv);
	}
}
