/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: net_send.c,v 1.13 2002/01/27 22:34:36 zarjazz Exp $
*/

#define __USE_SERVER_STRUCT
#include <ircd_include.h>

/* Prototypes */
client_t **	get_server_ids	(void);

/*
================
send_local
----------------
all outbound traffic goes via this function
================
*/

static void send_local (client_t *cl, byte *buffer, int len)
{
	sendq_pkt_t *entry;

	entry = sys_malloc(sizeof(*entry));
	entry->pkt.size = len;
	entry->pkt.data = malloc(len);
	assert(entry->pkt.data);
	memcpy(entry->pkt.data,buffer,len);

	sendq_insert(&cl->sendq,entry);
	/* only signal a thread to send if one isn't already busy with this client */
	if (!bit_tst(cl,FL_SENDING))
	{
		bit_set(cl,FL_SENDING);
		task_create(callback_client_send, (void *) cl->id, NULL);
	}
}

/*
================
send_remote
================
*/

static void send_remote (client_t *cl, byte *buffer, int len)
{
	client_t *route;
	byte msg[BUFSIZE];
	int msg_len;

	if (is_server(cl))
	{
		msg_len = sprintf(msg, "MSG %hu :", cl->node.id);
		route = server_route(cl);
	}
	else
	{
		client_t *sv = cl->local.server;

		msg_len = sprintf(msg, "MSG %s :", cl->name);
		if (is_local(sv))
			route = sv;
		else
			route = server_route(sv);
	}
	memcpy(msg + msg_len, buffer, len);
	msg_len += len;

	/* route this message via local server */
	send_local(route, msg, msg_len);
}

/*
================
cl_send_raw
================
*/

void cl_send_raw (client_t *cl, byte *buffer, int len)
{
	if (!cl)
	{
		u_int16_t i;
		client_t **servers = get_server_ids();
		/* broadcast to all neighbours */
		for (i = 0; i < Node->num_neighbours; i++)
			cl_send_raw(servers[Node->neighbours[i]], buffer, len);
	}
	else if (is_local(cl))
		send_local(cl, buffer, len);
	else
		send_remote(cl, buffer, len);
}

/*
================
chan_send_raw
================
*/

static void chan_send_raw (channel_t *chan, byte *buffer, int len,
					client_t *exclude, u_int32_t fl_cl, u_int32_t fl_index)
{
	u_int16_t list[NET_MAX_SERVERS], nservers = 0, i;
	byte table[NET_MAX_SERVERS];
	index_t *index;

	/* initialize server destination table */
	memset(table,0,sizeof(table));

	/* if FL_SERVER set, then broadcast msg (remember to clear the bit) */
	if (fl_cl & FL_BROADCAST)
	{
		cl_send_raw(NULL,buffer,len);
		fl_cl &= ~FL_BROADCAST;
	}

	/* send line to all channel members (possibly excluding client) */
	for (index = (index_t *) chan->members.item_start; index; index = index->next)
	{
		client_t *cl;

		cl = (client_t *) index->item;
		/* exclusions */
		if (exclude == cl) continue;
		if (fl_cl && !bit_tst(cl,fl_cl)) continue;
		if (fl_index && !bit_tst(index,fl_index)) continue;

		/* send to clients */
		if (is_local(cl))
			cl_send_raw(cl, buffer, len);
		else
		{
			u_int16_t id;
			/* skip remote clients not fully sync'd */
			if (bit_tst(cl,FL_BUSY)) continue;
			id = local_server_id(cl);
			if (!table[id])
			{
				table[id] = 1;
				list[nservers++] = id;
			}
		}
	}

	/* send text to remote servers */
	if (nservers)
	{
		client_t **servers = get_server_ids();
		byte msg[BUFSIZE];
		int msg_len;

		msg_len = sprintf(msg, "MSG %s :", chan->name);
		memcpy(msg + msg_len, buffer, len);
		msg_len += len;

		for (i = 0; i < nservers; i++)
			cl_send_raw(servers[list[i]], msg, msg_len);
	}
}

/*
================
cl_send_numeric
================
*/

void cl_send_numeric (client_t *cl, numeric_t numeric, ...)
{
	const char *rpl = numeric_string(numeric);
	byte buffer[BUFSIZE];
	va_list	argptr;
	int len;

	/* prepare network numeric string */
	if (is_server(cl))
		len = sprintf(buffer,"NOTICE %s : ", Srv->name);
	else
		len = sprintf(buffer,":%s %.3d %s ", Srv->name, (int) numeric, cl->name[0] ? cl->name : "*");

	va_start(argptr,numeric);
	vsnprintf(buffer + len, sizeof(buffer) - len - 2, rpl, argptr);
	va_end(argptr);

	/* remove spurious white spaces */
	len = strlen(buffer);
	while (len > 0 && isspace(buffer[len - 1])) buffer[--len] = '\0';

	/* append terminating newline */
	buffer[len++] = '\r';
	buffer[len++] = '\n';

	/* send line to client */
	cl_send_raw(cl,buffer,len);
}

/*
================
chan_send_numeric
================
*/

void chan_send_numeric (channel_t *chan, client_t *cl,
						client_t *exclude, numeric_t numeric, ...)
{
	const char *rpl = numeric_string(numeric);
	byte buffer[BUFSIZE];
	va_list	argptr;
	int len;

	/* prepare network numeric string */
	len = sprintf(buffer,":%s %.3d %s ", Srv->name, (int) numeric, cl->name[0] ? cl->name : "*");

	va_start(argptr,numeric);
	vsnprintf(buffer + len, sizeof(buffer) - len - 2, rpl, argptr);
	va_end(argptr);

	/* remove spurious white spaces */
	len = strlen(buffer);
	while (len > 0 && isspace(buffer[len - 1])) buffer[--len] = '\0';

	/* append terminating newline */
	buffer[len++] = '\r';
	buffer[len++] = '\n';

	/* send line to all channel members (possibly excluding client) */
	chan_send_raw(chan, buffer, len, exclude, 0, 0);
}

/*
================
cl_send_fmt
================
*/

void cl_send_fmt (client_t *cl, const char *fmt, ...)
{
	byte buffer[BUFSIZE];
	va_list	argptr;
	int len;

	/* prepare string */
	va_start(argptr,fmt);
	vsnprintf(buffer, sizeof(buffer) - 2, fmt, argptr);
	va_end(argptr);

	/* remove spurious white spaces */
	len = strlen(buffer);
	while (len > 0 && isspace(buffer[len - 1])) buffer[--len] = '\0';

	/* append terminating newline */
	buffer[len++] = '\r';
	buffer[len++] = '\n';

	/* send line to client */
	cl_send_raw(cl,buffer,len);
}

/*
================
chan_send_fmt
================
*/

void chan_send_fmt (channel_t *chan,
					client_t *exclude, u_int32_t fl_cl, u_int32_t fl_index,
					const char *fmt, ...)
{
	byte buffer[BUFSIZE];
	va_list	argptr;
	int len;

	assert(chan);
	/* prepare string */
	va_start(argptr,fmt);
	vsnprintf(buffer, sizeof(buffer) - 2, fmt, argptr);
	va_end(argptr);

	/* remove spurious white spaces */
	len = strlen(buffer);
	while (len > 0 && isspace(buffer[len - 1])) buffer[--len] = '\0';

	/* append terminating newline */
	buffer[len++] = '\r';
	buffer[len++] = '\n';

	/* send line to all channel members (possibly excluding client) */
	chan_send_raw(chan, buffer, len, exclude, fl_cl, fl_index);
}

/*
================
cl_send_names
================
*/

void cl_send_names (client_t *cl, channel_t *chan, bool show)
{
	index_t *index = NULL;
	byte buf[BUFSIZE];
	int len;

	/* send names to client */
	len = sprintf(buf,":%s %.3d %s %c %s :",
		Srv->name, (int) RPL_NAMREPLY, cl->name[0] ? cl->name : "*", bit_tst(chan, CM_SECRET) ? '@' : '=', chan->name);
	do
	{
		int n;
		n = channel_names(buf + len, sizeof(buf) - len, chan, &index, show);
		cl_send_raw(cl,buf,len + n);
	} while (index);
	cl_send_numeric(cl, RPL_ENDOFNAMES, chan->name);
}

/*
================
cl_send_njoin
cl_send_npart
================
*/

static void cl_send_nmsg (client_t *cl, const char *nmsg, char *list[], int j)
{
	char msg[BUFSIZE];
	unsigned len;
	int i;

	if (j)
	{
		len = sprintf(msg, nmsg, Srv->node.id, cl->name);
		for (i = 0; i < j; i++) len += sprintf(msg + len, " %s", list[i]);
		msg[len++] = '\r';
		msg[len++] = '\n';

		cl_send_raw(NULL, msg, len);
	}
}

void cl_send_njoin (client_t *cl, char *joined[], int j) { cl_send_nmsg(cl, ":%hu NJOIN %s", joined, j); }
void cl_send_npart (client_t *cl, char *parted[], int j) { cl_send_nmsg(cl, ":%hu NPART %s", parted, j); }
