/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Read the LICENSE file for more details
	----
	$Id: net_ssl.c,v 1.3 2002/01/13 21:33:36 zarjazz Exp $
*/


#include <ircd_include.h>

#ifdef HAVE_SSL

/* ================ */

SSL_CTX *Ctx;

/* SSL thread locking */
static mutex_t *lock_cs;
static long *lock_count;

static unsigned long thread_id (void)
{
	return (unsigned long) pthread_self();
}

static void locking_callback (int mode, int type, const char *file, int line)
{
	if (mode & CRYPTO_LOCK)
	{
		MUTEX_LOCK(lock_cs[type]);
		lock_count[type]++;
	}
	else
		MUTEX_UNLOCK(lock_cs[type]);
}

static void SSL_thread_init (void)
{
	int i;

	lock_cs = sys_malloc(CRYPTO_num_locks() * sizeof(mutex_t));
	lock_count = sys_malloc(CRYPTO_num_locks() * sizeof(long));

	for (i = 0; i < CRYPTO_num_locks(); i++)
	{
		lock_count[i] = 0;
		MUTEX_INIT(lock_cs[i]);
	}

	CRYPTO_set_id_callback(thread_id);
	CRYPTO_set_locking_callback(locking_callback);
}

static void SSL_thread_cleanup (void)
{
	int i;

	CRYPTO_set_locking_callback(NULL);

	for (i = 0; i < CRYPTO_num_locks(); i++) MUTEX_DESTROY(lock_cs[i]);

	sys_free(lock_cs);
	sys_free(lock_count);
}

/*
================
cleanup_ssl
================
*/

static void cleanup_ssl (void)
{
	SSL_thread_cleanup();
	SSL_CTX_free (Ctx);
}

/*
================
init_ssl
================
*/

void init_ssl (void)
{
	int rc;
/*	long opt; */
	SSL_METHOD *method;
	char *cert = ircd_config_string(Config, NULL, NULL, "SSLCertificate");

	SSL_load_error_strings();
	SSL_library_init();

	method = SSLv23_method();
	if (!method) sys_error("Failed to create SSL %s object", "method");

	Ctx = SSL_CTX_new(method);
	if (!Ctx) sys_error("Failed to create SSL %s object", "CTX");
/*
	opt = SSL_CTX_get_options(Ctx);
	opt |= SSL_OP_NO_TLSv1;
	opt |= SSL_OP_NO_SSLv2;
	SSL_CTX_set_options(Ctx, opt);
*/
	rc = SSL_CTX_use_certificate_file(Ctx, cert, SSL_FILETYPE_PEM);
	if (rc <= 0) sys_error("Failed to load SSL %s", "Certificate");

	rc = SSL_CTX_use_PrivateKey_file(Ctx, cert, SSL_FILETYPE_PEM);
	if (rc <= 0) sys_error("Failed to load SSL %s", "Private Key");
	
	rc = SSL_CTX_check_private_key(Ctx);
	if (!rc) sys_error("Private key does not match the certificate public key");

	rc = SSL_CTX_set_cipher_list(Ctx, "DEFAULT");
	if (rc <= 0) sys_error("SSL_CTX_set_cipher_list");

	SSL_thread_init();

	sys_color(s_green);
	sys_printf("SSL initialized\n");
	sys_color(s_normal);

	/* register for cleanup */
	sys_cleanup_register(cleanup_ssl);
}

/*
================
ssl_send
================
*/

int ssl_send (sys_pkt_t *pkt, SSL *ssl)
{
	int rc = SSL_write(ssl, pkt->data, pkt->size);

	/* write unsuccessful */
	if (rc <= 0)
	{
		struct timeval timeout;
		fd_set fds, *fdp = &fds;
		unsigned fd;
		int ret;

		ret = SSL_get_error(ssl, rc);
		rc = -1;

		/* setup the select */
		fd = SSL_get_fd(ssl);
		FD_ZERO(fdp);
		FD_SET(fd,fdp);

		/* 5 second timeout */
		timeout.tv_sec	= 5;
		timeout.tv_usec	= 0;

		/* well, this is an interesting "feature" of non-blocking SSL connections */
		switch (ret)
		{
		case SSL_ERROR_WANT_READ:
			ret = select(fd+1, fdp, NULL, NULL, &timeout);
			if (ret == 0)
				return -1; /* timeout */
			else if (ret > 0)
				return ssl_send(pkt,ssl);
			else if (ret < 0)
				sys_warn("ssl_send - read: %s", net_str_error());
			break;
		case SSL_ERROR_WANT_WRITE:
			ret = select(fd+1, NULL, fdp, NULL, &timeout);
			if (ret == 0)
				return -1; /* timeout */
			else if (ret > 0)
				return ssl_send(pkt,ssl);
			else if (ret < 0)
				sys_warn("ssl_send - write: %s", net_str_error());
			break;
		default:
			{
				unsigned long err;
				char msg[BUFSIZE];

				sys_warn("ssl_send: %d", ret);
				while (err = ERR_get_error()) sys_warn("  %s", ERR_error_string(err, msg));
			}
			break;
		}
	}
	else if ( sys_debug_get() )
	{
		struct sockaddr name;
		int namelen = sizeof(name);
		int sock = SSL_get_fd(ssl);

		getpeername(sock,&name,&namelen);
		net_dprintf(pkt,"send (%d) -> %s\n",&name);
	}

	return rc;
}

/*
================
ssl_read
================
*/

int ssl_read (sys_pkt_t *pkt, SSL *ssl, struct sockaddr *src, size_t len)
{
	pkt->size = SSL_read(ssl, pkt->data, len);

	/* read unsuccessful */
	if (pkt->size <= 0)
	{
		struct timeval timeout;
		fd_set fds, *fdp = &fds;
		unsigned fd;
		int ret;

		ret = SSL_get_error(ssl, pkt->size);
		pkt->size = -1;

		/* setup the select */
		fd = SSL_get_fd(ssl);
		FD_ZERO(fdp);
		FD_SET(fd,fdp);

		/* 5 second timeout */
		timeout.tv_sec	= 5;
		timeout.tv_usec	= 0;

		/* well, this is an interesting "feature" of non-blocking SSL connections */
		switch (ret)
		{
		case SSL_ERROR_WANT_READ:
			ret = select(fd+1, fdp, NULL, NULL, &timeout);
			if (ret == 0)
				return -1; /* timeout */
			else if (ret > 0)
				return ssl_read(pkt,ssl,src,len);
			else if (ret < 0)
				sys_warn("ssl_read - read: %s", net_str_error());
			break;
		case SSL_ERROR_WANT_WRITE:
			ret = select(fd+1, NULL, fdp, NULL, &timeout);
			if (ret == 0)
				return -1; /* timeout */
			else if (ret > 0)
				return ssl_read(pkt,ssl,src,len);
			else if (ret < 0)
				sys_warn("ssl_read - write: %s", net_str_error());
			break;
		default:
			{
				unsigned long err;
				char msg[BUFSIZE];

				sys_warn("ssl_read: %d", ret);
				while (err = ERR_get_error()) sys_warn("  %s", ERR_error_string(err, msg));
			}
			break;
		}
	}
	else if ( sys_debug_get() )
		net_dprintf(pkt,"read (%d) <- %s\n",src);

	return pkt->size;
}

#endif /* HAVE_SSL */


