/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Please read the LICENSE file for more details
	----
	$Id: sys_common.c,v 1.10 2001/12/14 20:52:07 zarjazz Exp $
*/

#include <sys_include.h>
#include <sys_common.h>
#include <sys_string.h>

#include <sys/stat.h>
#include <time.h>

#ifdef _REENTRANT
  #include <pthread.h>
#endif

/* ANSI Color Terminal Settings */

char s_red[]	= { 0x1b,'[','1',';','3','1','m','\0' };
char s_green[]	= { 0x1b,'[','1',';','3','2','m','\0' };
char s_yellow[]	= { 0x1b,'[','1',';','3','3','m','\0' };
char s_purple[] = { 0x1b,'[','1',';','3','5','m','\0' };
char s_cyan[]	= { 0x1b,'[','1',';','3','6','m','\0' };
char s_normal[]	= { 0x1b,'[','0',';','3','9','m','\0' };

void sys_color (char *s)
{
#ifndef WINVER
	fputs(s,stdout);
	fflush(stdout);
#endif
}

/* Locals */

static int	_debugLevel	= 0;

/* (Logged) I/O */

static int			_logMode		= 1;
static const char *	_logFilename	= NULL;
static FILE *		_logFile		= NULL;

#ifdef _REENTRANT
  pthread_mutex_t	_logMutex		= PTHREAD_MUTEX_INITIALIZER;
#endif

/* cleanup function table */

static int _cleanup_table_size = 0;
static cleanup_func_t *_cleanup_table = NULL;

/*
================
sys_cleanup(_register)
================
*/

void sys_cleanup_register (cleanup_func_t func)
{
	_cleanup_table = sys_realloc(_cleanup_table, sizeof(cleanup_func_t *) * (_cleanup_table_size + 1));
	_cleanup_table[_cleanup_table_size++] = func;
}

void sys_cleanup (void)
{
	while (_cleanup_table_size--) _cleanup_table[_cleanup_table_size]();
	sys_free(_cleanup_table);
}

/*
================
sys_debug_get/set
================
*/

int  sys_debug_get (void)    { return _debugLevel; }
void sys_debug_set (int lvl) { _debugLevel = lvl; }

/*
================
sys_str_error
================
*/

char *sys_str_error (void)
{
#ifdef WINVER
	return strerror( errno );
#else
	return strerror( errno );
#endif
}

/*
================
sys_str_time
----------------
returns a string with the local time
in "dd:mm:yy" format
================
*/

const char *sys_str_time (void)
{
	struct tm *newtime;
	time_t tm;
	char *s;

	time( &tm );
	newtime = localtime( &tm );
	s = asctime(newtime);
	/* return string with newline removed */
	return str_strip(s);
}

/*
================
sys_dprintf
================
*/

void sys_dprintf (int lvl, const char *msg, ...)
{
	va_list	argptr;

#ifndef _DEBUG
	if (_debugLevel < abs(lvl)) return;
#endif

	va_start(argptr,msg);

	/* FIXME: why does removing this line break everything?! */
	vprintf(msg,argptr);

	if (lvl >= 0 && _logFile)
	{
#ifdef _REENTRANT
		pthread_mutex_lock(&_logMutex);
#endif
		vfprintf(_logFile,msg,argptr);
		fflush(_logFile);
#ifdef _REENTRANT
		pthread_mutex_unlock(&_logMutex);
#endif
	}
	va_end(argptr);
}

/*
================
sys_printf
================
*/

void sys_printf (const char *msg, ...)
{
	va_list	argptr;

	va_start(argptr,msg);
	vprintf(msg,argptr);
	if (_logFile)
	{
#ifdef _REENTRANT
		pthread_mutex_lock(&_logMutex);
#endif
		vfprintf(_logFile,msg,argptr);
		fflush(_logFile);
#ifdef _REENTRANT
		pthread_mutex_unlock(&_logMutex);
#endif
	}
	va_end(argptr);
}

/*
================
sys_error
----------------
prints an error message and exit process
================
*/

void sys_error (const char *msg, ...)
{
	va_list argptr;

	va_start(argptr,msg);
	fputs("** Error: ",stderr);
	vfprintf(stderr,msg,argptr);
	fputs("\n",stderr);
	if (_logFile)
	{
#ifdef _REENTRANT
		pthread_mutex_lock(&_logMutex);
#endif
		fputs("** Error: ",_logFile);
		vfprintf(_logFile,msg,argptr);
		fputs("\n",_logFile);
#ifdef _REENTRANT
		pthread_mutex_unlock(&_logMutex);
#endif
	}
	va_end(argptr);

	exit(EXIT_FAILURE);
}

/*
================
sys_warn
----------------
prints a warning message
================
*/

void sys_warn (const char *msg, ...)
{
	va_list argptr;

	va_start(argptr,msg);
	fputs("** Warning: ",stderr);
	vfprintf(stderr,msg,argptr);
	fputs("\n",stderr);
	if (_logFile)
	{
#ifdef _REENTRANT
		pthread_mutex_lock(&_logMutex);
#endif
		fputs("** Warning: ",_logFile);
		vfprintf(_logFile,msg,argptr);
		fputs("\n",_logFile);
#ifdef _REENTRANT
		pthread_mutex_unlock(&_logMutex);
#endif
	}
	va_end(argptr);
}

/*
================
sys_print_raw
----------------
print data to a file in hexdump format
================
*/

#define COL_WIDTH	8
#define COL_COUNT	2

void sys_print_raw (FILE *f, byte *d, size_t size)
{
	unsigned int i,j,n,x;
	byte *c;
	int k;

	assert(f);

	while (size)
	{
		/* Hex Output */
		c = d;
		x = size;
		fputs("  ",f);
		for (k = 0; k < COL_COUNT; k++)
		{
			if (x >= COL_WIDTH)
				{ i = COL_WIDTH; j = 0; }
			else
				{ i = x; j = COL_WIDTH - x; }
			for (n=0; n < i; n++) fprintf(f,"%.2x ",(int) *c++);
			for (n=0; n < j; n++) fputs("   ",f);
			fputc(' ',f);
			x -= i;
		}
		/* Ascii Output */
		c = d;
		x = size;
		fputc(' ',f);
		for (k = 0; k < COL_COUNT; k++)
		{
			if (x >= COL_WIDTH)
				{ i = COL_WIDTH; j = 0; }
			else
				{ i = x; j = COL_WIDTH - x; }
			for (n=0; n < i; n++)
			{
				byte a;
				a = *c++;
				fputc(isprint(a) ? a : '.',f);
			}
			for (n=0; n < j; n++) fputc(' ',f);
			x -= i;
		}
		/* End */
		d = c;
		size = x;
		fputs("\n",f);
	}
}

/*
================
sys_log_init
----------------
mode: 0 = none, 1 = new, 2 = append
================
*/

void sys_log_init (int mode, const char *fname)
{
	char *fmode;

	/* register for cleanup */
	sys_cleanup_register(sys_log_cleanup);

	/* sanity check */
	if (!mode || !fname) return;
	fmode = (mode == 1) ? "w" : "a";

	_logMode = mode;
	_logFile = fopen(fname,fmode);
	_logFilename = fname;

	if (!_logFile) sys_warn("Unable to open log file: %s\n",fname);
}

/*
================
sys_log_cleanup
----------------
close the log file
================
*/

void sys_log_cleanup (void)
{
	if (_logFile) fclose(_logFile);
	_logFile = NULL;
}

/*
================
sys_log_get
================
*/

FILE *sys_log_get (void) { return _logFile; }

/*
================
sys_file_read
----------------
read an entire file to a malloc'd buffer and ensure
buffer is null terminated (for string functions)
================
*/

byte *sys_file_read (const char *fname, size_t *size)
{
	struct stat filestats;
	size_t fsize;
	byte *fbuf;
	FILE *f;
	int rc;

	f = fopen(fname,"rb");
	if (!f)	{ sys_warn("Unable to open file: %s",fname); return NULL; }

	rc = fstat(fileno(f),&filestats);
	if (rc < 0) { sys_error("fstat: %s", sys_str_error()); return NULL; }

	fsize = filestats.st_size;
	if (size) *size = fsize;
	if (!fsize) return NULL;

	fbuf = (byte *) sys_malloc(fsize + 1);
	fread(fbuf,fsize,1,f);
	fclose(f);

	fbuf[fsize] = '\0';

	return fbuf;
}

/* Memory (De)Allocation */

void *sys_malloc (size_t size)
{
	void *ptr = (void *) malloc(size);
	if (!ptr) { sys_error("malloc (%u) bytes",size); return NULL; }
	memset(ptr,0,size);
	return ptr;
}

void *sys_realloc (void *ptr, size_t size)
{
	ptr = (void *) realloc(ptr,size);
	if (!ptr) sys_error("realloc (%u) bytes",size);
	return ptr;
}

void sys_free (void *ptr) { if (ptr) free(ptr); }

#ifdef WINVER

void *win_malloc (size_t size)
{
	void *ptr = (void *) GlobalAllocPtr(GMEM_MOVEABLE | GMEM_SHARE,size);
	if (!ptr) { sys_error("win_malloc (%u) bytes",size); return NULL; }
	memset(ptr,0,size);
	return ptr;
}

void *win_realloc (void *ptr, size_t size)
{
	if (ptr)
		ptr = GlobalReAllocPtr(ptr,size,GMEM_MOVEABLE | GMEM_SHARE);
	else
		ptr = GlobalAllocPtr(GMEM_MOVEABLE | GMEM_SHARE,size);
	if (!ptr) sys_error("win_realloc (%u) bytes",size);
	return ptr;
}

void win_free (void *ptr) { if (ptr) GlobalFreePtr(ptr); }

#endif /* WINVER */
