/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Please read the LICENSE file for more details
	----
	$Id: sys_config_block.c,v 1.2 2001/11/30 16:26:19 zarjazz Exp $
*/

#include <sys_include.h>
#include <sys_common.h>
#include <sys_string.h>
#include <sys_array.h>
#include <sys_list.h>
#include <sys_config.h>

#ifdef WINVER
  #pragma warning(disable : 4090)	// const
#endif

static void parse_block_body (block_t *global, block_t *block, char *buffer);

/*
================
init_block
----------------
parse config file
================
*/

block_t *init_block (const char *fname)
{
	block_t *config = NULL;
	u_int32_t size;
	byte *buf;

	buf = sys_file_read(fname, &size);
	if (!buf) goto bail;

	config = sys_malloc(sizeof(*config));
	config->block_name = str_dup("Root");

	parse_block_body(config,config,buf);

bail:
	sys_free(buf);
	return config;
}

/*
================
cleanup_block_arg
================
*/

static void cleanup_block_arg (block_arg_t *val)
{
	sys_free(val->arg_value);
	sys_free(val->arg_name);
}

/*
================
cleanup_block
----------------
free() a config block & all sub-blocks
================
*/

void cleanup_block (block_t *block)
{
	int i;
	block_t *ptr, *prev;
	block_arg_t **args;

	if (!block) return;

	for (ptr = (block_t *) block->sub_blocks.item_end; ptr; ptr = prev)
	{ 
		prev = ptr->prev;
		cleanup_block(ptr);
	}

	args = block->block_args;
	for (i = 0; i < block->nargs; i++) cleanup_block_arg(args[i]);

	sys_free(block->block_value);
	sys_free(block->block_name);
	sys_free(block);
}

/*
================
print_block
----------------
print config block & all sub-blocks
================
*/

#define INDENT(x)	{ int n = x; while (n--) sys_printf("  "); }
 
void print_block (block_t *block, int lvl)
{
	int i;
	block_t *ptr;
	block_arg_t **args = block->block_args;

	INDENT(lvl);
	if (block->block_value)
		sys_dprintf(1,"<%s %s>\n",block->block_name, block->block_value);
	else
		sys_dprintf(1,"<%s>\n",block->block_name);

	for (i = 0; i < block->nargs; i++)
	{
		block_arg_t *arg;

		INDENT(lvl+1);
		arg = args[i];
		if (arg->arg_value)
			sys_dprintf(1,"%s %s\n",arg->arg_name, arg->arg_value);
		else
			sys_dprintf(1,"%s\n",arg->arg_name);
	}

	for (ptr = (block_t *) block->sub_blocks.item_start; ptr; ptr = ptr->next)
		print_block(ptr,lvl+1);

	INDENT(lvl);
	sys_dprintf(1,"</%s>\n",block->block_name);
}

/* ================ */
/*  Parse Routines  */
/* ================ */

static char *skip_space (const char *src)
	{ while (*src && isspace(*src)) src++; return (char *) src; }
static char *skip_alpha (const char *src)
	{ while (*src && isalpha(*src)) src++; return (char *) src; }
static char *skip_alnum (const char *src)
	{ while (*src && isalnum(*src)) src++; return (char *) src; }

static char *str_nextline (const char *src)
{
	return (char *) strchr(src,'\n');
}

static char *str_token (char *dst, const char *src, int dst_len)
{
	src = skip_space(src);

	if (*src == '"')
	{
		src++;
		while (*src && (--dst_len > 0) && (*src != '"')) *dst++ = *src++;
	}
	else
		while (*src && (--dst_len > 0) && !isspace(*src)) *dst++ = *src++;
	*dst = '\0';

	return (char *) src;
}

/*
================
parse_block_header
----------------
parse the <..> block </..>
================
*/

static char *parse_block_header (block_t *global, block_t *current, char *src)
{
	block_t *block = NULL;
	char *str, *end, *ret, *err, buf[256];

	/* parse header */
	str = src + 1;
	if ( !isalpha(*str) ) { err = "< >"; goto bail; }

	end = strchr(str,'>');
	if (!end) { err = "missing '>' termintor"; goto bail; }
	*end++ = '\0';

	block = sys_malloc(sizeof(*block));

	/* block name */
	str = str_token(buf, str, sizeof(buf));
	if ( !strlen(buf) ) { err = "null block name"; goto bail; }

	block->block_name = str_dup(buf);

	/* block argument */
	str = str_token(buf, str, sizeof(buf));
	if (buf[0]) block->block_value = str_dup(buf);

	/* find terminator block */
	str_copy(buf,"</",sizeof(buf));
	str_copy(buf+2,block->block_name,sizeof(buf) - 2);
	strcat(buf,">");

	str = strstr(end,buf);
	if (!str) { err = buf; goto bail; }

	ret = str + strlen(buf) + 1;
	*str = '\0';

	/* add sub-block to current block */
	ll_add_end(&current->sub_blocks,block);

	/* parse the block */
	parse_block_body(global,block,end);

	return ret;
bail:
	cleanup_block(block);
	sys_error("Syntax Error: %s", err);
	return NULL;
}

/*
================
parse_arguments
----------------
parse the argument line
================
*/

static char *parse_arguments (block_t *block, char *src)
{
	char *str, *ret, buf[256];
	block_arg_t *arg;

	str = src;
	ret = str_nextline(str);
	if (ret) *ret++ = '\0';

	/* resize argument array */
	block->block_args = sys_realloc(block->block_args, sizeof(block_arg_t *) * (block->nargs + 1));
	block->block_args[block->nargs++] = arg = sys_malloc(sizeof(block_arg_t));

	/* extract argument name */
	str = str_token(buf, str, sizeof(buf));
	if (!buf[0]) return ret;
	arg->arg_name = str_dup(buf);

	/* extract argument value */
	str = str_token(buf, str, sizeof(buf));
	if (buf[0]) arg->arg_value = str_dup(buf);

	return ret;
}

/*
================
parse_block_body
================
*/

static void parse_block_body (block_t *global, block_t *block, char *buffer)
{
	char *str = buffer;

	/* Begin */
	while ( str && *(str = skip_space(str)) )
	{
		/* Comment */
		if ( str[0] == '#' || (str[0] == '/' && str[1] == '/') )
			str = str_nextline(str);
		/* Section */
		else if ( str[0] == '<' )
			str = parse_block_header(global, block, str);
		/* Argument */
		else if ( isascii(str[0]) )
			str = parse_arguments(block, str);
		/* Syntax Error */
		else
		{
			sys_error("Syntax Error: ");
			return;
		}
	}
}

/*
================
find_block
================
*/

static conf_block_t *find_block (conf_block_t **table, int table_size, const char *name)
{
	conf_block_t *conf_block;
	int i;

	for (i = 0; i < table_size; i++)
	{
		conf_block = table[i];
		if ( !stricmp(conf_block->name,name) ) return conf_block;
	}
	return NULL;
}

/*
================
find_entry
================
*/

static conf_entry_t *find_entry (conf_block_t *conf_block, const char *name)
{
	conf_entry_t *entry = conf_block->entries;

	while (entry->value)
	{
		if ( !stricmp(entry->value, name) ) return entry;
		entry++;
	}
	return NULL;
}

/*
================
block_syntax_check
================
*/

void block_syntax_check (block_t *block, conf_block_t **block_table, int block_table_size)
{
	block_arg_t **args = block->block_args;
	conf_block_t *conf_block;
	conf_entry_t *conf_entry;
	char *err = NULL, *err_arg = "";
	block_t *ptr;
	int i;

	/* sanity check block name */
	conf_block = find_block(block_table,block_table_size,block->block_name);
	if (!conf_block) { err = "illegal block name"; goto bail; }

	/* sanity check argument entries */
	for (i = 0; i < block->nargs; i++)
	{
		block_arg_t *arg;

		arg = args[i];
		conf_entry = find_entry(conf_block,arg->arg_name);
		if (!conf_entry) { err = "illegal entry -"; err_arg = arg->arg_name; goto bail; }
	}

	/* sanity check argument flags */
	for (conf_entry = conf_block->entries; conf_entry->value; conf_entry++)
	{
		if (conf_entry->flags & CF_REQUIRED)
		{
			bool found = false;
			for (i = 0; i < block->nargs; i++) if ( !stricmp(conf_entry->value, args[i]->arg_name) )
				{ found = true; break; }
			if (!found) { err = "missing value -"; err_arg = conf_entry->value; goto bail; }
		}

		if ( !(conf_entry->flags & CF_MULTIREF) )
		{
			int count = 0;

			for (i = 0; i < block->nargs; i++) if ( !stricmp(conf_entry->value, args[i]->arg_name) )
				count++;
			if (count > 1) { err = "multiple values -"; err_arg = conf_entry->value; goto bail; }
		}
	}

	/* sanity check sub blocks */
	for (ptr = (block_t *) block->sub_blocks.item_start; ptr; ptr = ptr->next)
		block_syntax_check(ptr,block_table,block_table_size);

bail:
	if (err) sys_error("block <%s>: %s %s",block->block_name,err,err_arg);
	return;
}

/*
================
blocks_by_name
----------------
used to loop over all <NAME> .. </NAME> blocks 
================
*/

block_t *blocks_by_name (block_t *block, block_t *previous, const char *name)
{
	block_t *ptr;

	for (ptr = (block_t *) block->sub_blocks.item_start; ptr; ptr = ptr->next)
		if ( !stricmp(ptr->block_name,name) )
		{
			if (!previous) return ptr;
			if (ptr == previous) previous = NULL;
		}

	return NULL;
}

/*
================
arguments_by_name
----------------
used to loop over all 'NAME' arguments within a block
================
*/

block_arg_t *arguments_by_name (block_t *block, block_arg_t *previous, const char *name)
{
	block_arg_t *vals, **args = block->block_args;
	int i;

	for (i = 0; i < block->nargs; i++)
	{
		vals = args[i];
		if ( !stricmp(vals->arg_name,name) )
		{
			if (!previous) return vals;
			if (vals == previous) previous = NULL;
		}
	}

	return NULL;
}

/*
================
argument_by_name
================
*/

char *argument_by_name (block_t *block, const char *name)
{
	block_arg_t *blk_arg;
	
	blk_arg = arguments_by_name(block, NULL, name);
	if (blk_arg) return blk_arg->arg_value;

	return NULL;
}

/*
================
argument_range
----------------
convert a range argument to an array of integer values
================
*/

void argument_range (block_arg_t *range, array_t *array)
{
	int i, a = 0, b = 0;
	char arg[512], *dst, *src = range->arg_value;
	bool do_range = false;

	/* convert range buffer to values */
	if (src) while (*src)
	{
		dst = arg;
		src = skip_space(src);

		if ( isdigit(*src) )
		{
			while (*src && isdigit(*src)) *dst++ = *src++;
			*dst = '\0';
			a = atoi(arg);

			/* add to argument array, skipping any values we have already added */
			if (do_range)
			{
				i = b - a;
				if (!i) sys_error("null range: %s", range->arg_value);

				if (i < 0)	/* a > b */
					for (i = b + 1; i <= a; i++) array_add_item(array,&i,sizeof(int));
				else		/* a < b */
					for (i = b - 1; i >= a; i--) array_add_item(array,&i,sizeof(int));
				do_range = false;
			}
			else
				array_add_item(array,&a,sizeof(int));
		}
		else if ( *src == '-' )
		{
			src++;
			b = a;
			do_range = true;
		}
		else if ( *src != '\0' )
		{
			sys_error("range syntax error: %s", range->arg_value);
			return;
		}
	}

	if (do_range) sys_error("missing range value: %s", range->arg_value);
}

/*
================
argument_list
----------------
convert a list argument to an array of strings
================
*/

void argument_list (block_arg_t *list, array_t *array)
{
	char arg[512], *dst, *src = list->arg_value;

	/* convert list buffer to values */
	if (src) while (*src)
	{
		src = str_extract_alnum(arg,src,sizeof(arg));
		if (arg[0])
		{
			dst = str_dup(arg);
			array_add_item(array, &dst, sizeof(char *));
		}
	}
}
