/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Please read the LICENSE file for more details
	----
	$Id: sys_network.c,v 1.18 2001/11/27 17:32:03 zarjazz Exp $
*/

#include <sys_include.h>
#include <sys_common.h>
#include <sys_string.h>
#include <sys_network.h>

#ifdef WINVER
	#include <sys/timeb.h>
#endif

#ifdef _REENTRANT
  #include <pthread.h>

  extern pthread_mutex_t _logMutex;
#endif

/*
===============
gettimeofday
===============
*/

#ifdef WINVER

void gettimeofday (struct timeval *tv, struct timeval *null)
{
	struct _timeb tb;

	_ftime(&tb);

	tv->tv_sec  = tb.time;
	tv->tv_usec = (unsigned int) tb.millitm * 1000;
}

#endif /* WINVER */

int32_t tv_diff (struct timeval *start, struct timeval *end)
{
	return (end->tv_sec - start->tv_sec) * 1000000 + end->tv_usec - start->tv_usec;
}

/*
================
init_net
================
*/

void init_net (void)
{
#ifdef WINVER
	static WSADATA ws;

	if (WSAStartup(MAKEWORD(1,1), &ws))
		sys_error("Winsock failure: %s", net_str_error());
#endif

	/* register for cleanup */
	sys_cleanup_register(cleanup_net);
}

/*
================
cleanup_net
================
*/

void cleanup_net (void)
{
#ifdef WINVER
	WSACleanup();
#endif
}

/*
================
net_str_error
================
*/

char *net_str_error (void)
{
#ifdef WINVER
	return strerror( WSAGetLastError() );
#else
	return strerror( errno );
#endif
}

/*
================
net_str_sockaddr
================
*/

char *net_str_sockaddr (char *s, struct sockaddr_in *sa)
{
#ifndef WINVER
	char ip[16 + 6];
#endif /* WINVER */

	assert(s);

#ifdef WINVER
	if (sa->sin_port)
		sprintf(s, "%s:%hu", inet_ntoa(sa->sin_addr), ntohs(sa->sin_port));
	else
		strcpy(s, inet_ntoa(sa->sin_addr));
#else
	inet_ntop(AF_INET, &sa->sin_addr, ip, sizeof(ip));

	if (sa->sin_port)
		sprintf(s, "%s:%hu", ip, ntohs(sa->sin_port));
	else
		strcpy(s, ip);
#endif /* WINVER */

	return s;
}

/*
================
net_str_local
================
*/

char *net_str_local (char *s)
{
	int rc;
	struct sockaddr_in sa;
	char host[MAXHOSTNAMELEN];
	// Get Local Interface Address
	assert(s);

	rc = gethostname(host,sizeof(host));
	if (rc < 0)
	{
		sys_warn("net_str_local: %s", net_str_error());
		strcpy(s,"127.0.0.1");
	}
	else
	{
		// Turn the Address into an IP
		net_sockaddr(&sa,host,0);
		net_str_sockaddr(s,&sa);
	}

	return s;
}

/*
================
net_cmp_sockaddr
================
*/

int net_cmp_sockaddr (struct sockaddr_in *a, struct sockaddr_in *b)
{
	return (a->sin_addr.s_addr == b->sin_addr.s_addr && a->sin_port == b->sin_port);
}

/*
================
net_resolve
----------------
converts an internet address in either IP dot format
or hostname, to in_addr format
================
*/

bool net_resolve (const char *host_name, struct in_addr *addr)
{
	if ( (addr->s_addr = inet_addr(host_name)) == INADDR_NONE )
	{
		struct hostent *host_ent = gethostbyname(host_name);
		if (host_ent)
			memcpy (&addr->s_addr, host_ent->h_addr, host_ent->h_length);
		else
			return true;
	}
	return false;
}

/*
================
net_sockaddr
----------------
initializes a 'sockaddr_in' structure given a host address and port #
================
*/

bool net_sockaddr (struct sockaddr_in *sa, const char *host_name, u_int16_t port)
{
	memset(sa,0,sizeof(*sa));
	sa->sin_family = AF_INET;
	sa->sin_port = htons(port);
	if (host_name)
		return net_resolve(host_name,&sa->sin_addr);
	else
		sa->sin_addr.s_addr = htonl(INADDR_ANY);
	return false;
}

/*
================
net_sock_noblock
----------------
place a socket into non-blocking mode
================
*/

int net_sock_noblock (u_int16_t sock)
{
	int rc;
#ifdef WINVER
	u_long mode = 1;
	rc = ioctlsocket(sock, FIONBIO, &mode);
#else
	rc = fcntl(sock, F_SETFL, O_NONBLOCK);
#endif
	if (rc < 0) sys_dprintf(1,"sock_noblock: %s\n", net_str_error());

	return rc;
}

/*
================
net_sock_nodelay
================
*/

int net_sock_nodelay (u_int16_t sock)
{
	int rc;
	const int yes = 1;

	rc = setsockopt(sock, SOL_SOCKET, TCP_NODELAY, (void *) &yes, sizeof(int));
	if (rc < 0) sys_dprintf(1,"sock_nodelay: %s\n", net_str_error());

	return rc;
}

/*
================
net_sock_buffer
================
*/

int net_sock_buffer (u_int16_t sock, unsigned int size)
{
	unsigned int opt = size;
	int rc;

	rc = setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (const char *) &opt, sizeof(opt));
	if (rc < 0)
	{
		sys_dprintf(1,"sock_buffer (%s): %s\n", "RCV", net_str_error());
		return rc;
	}

	rc = setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (const char *) &opt, sizeof(opt));
	if (rc < 0) sys_dprintf(1,"sock_buffer (%s): %s\n", "SND", net_str_error());

	return rc;
}

/*
================
net_sock_udp
----------------
create & bind a UDP socket to given ip:port
================
*/

int net_sock_udp (struct sockaddr_in *name)
{
	int sock;

	/* create the UDP socket */
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) { sys_warn("socket(): udp"); return -1; }

	/* bind socket */
	if (bind(sock, (struct sockaddr *) name, sizeof(*name)) < 0) {sys_warn("bind(): udp"); return -1; }

	return sock;
}

/*
================
net_sock_tcp
----------------
create & bind a TCP socket to given ip:port
================
*/

int net_sock_tcp (struct sockaddr_in *name)
{
	int sock, flags = 1;
	struct linger lin = { 1, 5 };

	/* create the TCP socket */
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) { sys_warn("socket(): tcp"); return -1; }


	/* bind socket */
	if (bind(sock, (struct sockaddr *) name, sizeof(*name)) < 0) { sys_warn("bind(): tcp"); return -1; }

	/* set tcp socket options */
	setsockopt(sock, SOL_SOCKET, SO_LINGER, (void *) &lin, sizeof(lin));
	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *) &flags, sizeof(flags));
	setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, (void *) &flags, sizeof(flags));

	return sock;
}

/*
================
net_sock_close
================
*/

void net_sock_close (u_int16_t sock)
{
	if (sock)
	{
#ifdef WINVER
		int rc = closesocket(sock);
#else
		int rc = close(sock);
#endif
		if (rc < 0) sys_dprintf(3,"sock_close: %s\n", net_str_error());
	}
}

/*
================
net_dprintf
================
*/

void net_dprintf (sys_pkt_t *pkt, char *str, struct sockaddr *name)
{
	char s[64];
	FILE *f = sys_log_get();

	if (!f) return;

#ifdef _REENTRANT
	pthread_mutex_lock(&_logMutex);
#endif

	fprintf(f, str, pkt->size, net_str_sockaddr(s,(struct sockaddr_in *) name));
	fprintf(f, "{\n");
	sys_print_raw(f,pkt->data,pkt->size);
	fprintf(f, "}\n");
	fflush(f);

#ifdef _REENTRANT
	pthread_mutex_unlock(&_logMutex);
#endif
}

/*
================
net_send
================
*/

int net_send (sys_pkt_t *pkt, u_int16_t sock)
{
	int rc = send(sock,pkt->data,pkt->size,0);

	while (rc == EAGAIN || (rc == -1 && errno == EAGAIN))
	{
		struct timeval timeout;
		fd_set fds, *fdp = &fds;
		unsigned fd = sock;

		/* setup the select */
		FD_ZERO(fdp);
		FD_SET(fd,fdp);

		/* 5 second timeout */
		timeout.tv_sec	= 5;
		timeout.tv_usec	= 0;

		/* wait until we write */
		rc = select(fd+1, NULL, fdp, NULL, &timeout);

		if (rc == 0) return -1; /* timeout */
		if (rc > 0) rc = send(sock,pkt->data,pkt->size,0);
	}

	if (rc < 0)
		sys_warn("net_send (%d) %s\n", errno, net_str_error());
	else if ( sys_debug_get() )
	{
		struct sockaddr name;
		int namelen = sizeof(name);

		getpeername(sock,&name,&namelen);
		net_dprintf(pkt,"send (%d) -> %s\n",&name);
	}

	return rc;
}

/*
================
net_sendto
================
*/

int net_sendto (sys_pkt_t *pkt, u_int16_t sock, struct sockaddr *dst)
{
	int rc = sendto(sock,pkt->data,pkt->size,0,dst,sizeof(*dst));
	
	while (rc == EAGAIN || (rc == -1 && errno == EAGAIN))
	{
		struct timeval timeout;
		fd_set fds, *fdp = &fds;
		unsigned fd = sock;

		/* setup the select */
		FD_ZERO(fdp);
		FD_SET(fd,fdp);

		/* 5 second timeout */
		timeout.tv_sec	= 5;
		timeout.tv_usec	= 0;

		/* wait until we write */
		rc = select(fd+1, NULL, fdp, NULL, &timeout);

		if (rc == 0) return -1; /* timeout */
		if (rc > 0) rc = sendto(sock,pkt->data,pkt->size,0,dst,sizeof(*dst));
	}

	if (rc < 0)
		sys_warn("net_sendto (%d) %s\n", errno, net_str_error());
	else if ( sys_debug_get() )
		net_dprintf(pkt,"send (%d) -> %s\n",dst);

	return rc;
}

/*
================
net_read
----------------
read packet data
================
*/

int net_read (sys_pkt_t *pkt, u_int16_t sock, struct sockaddr *src, size_t len)
{
	pkt->size = recv(sock,pkt->data,len,0);
	if (pkt->size < 0)
		sys_warn("net_read - %s\n", net_str_error());
	else if ( sys_debug_get() )
		net_dprintf(pkt,"read (%d) <- %s\n",src);

	return pkt->size;
}

