/*
	Author	: Zarjazz
	Email	: zarjazz@sircd.org
	WebPage	: http://www.sircd.org/

	Copyright (C) 2001,2002  Vincent Sweeney
	----
	Please read the LICENSE file for more details
	----
	$Id: sys_string.c,v 1.8 2001/11/30 16:26:19 zarjazz Exp $
*/

#include <sys_include.h>
#include <sys_common.h>
#include <sys_string.h>

#ifdef WINVER
  #pragma warning(disable : 4090)	// const
#endif

/*
================
str_strip
----------------
remove initial and trailing white-space
char's from a string
================
*/

char *str_strip (char *str)
{
	int l;

	assert(str);
	/* skip pre-non-valid chars */
	while (*str && isspace(*str)) str++;
	/* remove trailing white space chars */
	l = strlen(str);
	while (l > 0 && isspace(str[--l])) str[l] = '\0';
	/* return the stripped string */
	return str;
}

/*
================
str_copy
----------------
buffer safe string copy
================
*/

char *str_copy (char *dst, const char *src, int dst_len)
{
	char *save = dst;

	assert(src);
	assert(dst);
	/* copy string and ensure null terminated */
	while (*src && (--dst_len > 0)) *dst++ = *src++;
	*dst = '\0';

	return save;
}

/*
================
str_extract
----------------
extract the first non-space token from 'src' and
ensure string is null terminated. returns the
string postition after extracted token
================
*/

char *str_extract (char *dst, const char *src, int dst_len)
{
	assert(src);
	assert(dst);
	/* skip pre-non-valid chars */
	while (*src && isspace(*src)) src++;
	/* copy string and ensure null terminated */
	while (*src && (--dst_len > 0) && !isspace(*src)) *dst++ = *src++;
	/* skip post-non-valid chars */
	while (*src && isspace(*src)) src++;
	*dst = '\0';

	return (char *) src;
}

/*
================
str_extract_valid
----------------
extract the first token of 'valid' characters from 'src'
================
*/

char *str_extract_valid (char *dst, const char *src, int dst_len, int (*func) (int))
{
	assert(src);
	assert(dst);
	/* skip pre-non-valid chars */
	while (*src && !func(*src)) src++;
	/* copy string and ensure null terminated */
	while (*src && (--dst_len > 0) && func(*src)) *dst++ = *src++;
	/* skip post-non-valid chars */
	while (*src && !func(*src)) src++;
	*dst = '\0';

	return (char *) src;
}

char *str_extract_alnum (char *dst, const char *src, int dst_len)
	{ return str_extract_valid(dst,src,dst_len,isalnum); }
char *str_extract_graph (char *dst, const char *src, int dst_len)
	{ return str_extract_valid(dst,src,dst_len,isgraph); }

/*
================
str_dup
----------------
duplicate a string
================
*/

char *str_dup (const char *src)
{
	int len; 
	char *dst;

	assert(src);
	/* malloc + copy string */
	len = strlen(src) + 1;
	dst = sys_malloc(len);
	if (!dst) return "";
	str_copy(dst,src,len);

	return dst;
}

/*
================
str_match
----------------
does a string match using a very simple "regex" (ie only special characters are: "*?\")
================
*/

bool str_match (const char *str, const char *reg, bool case_sensitive)
{
	assert(str);
	assert(reg);
	/* compare strings */
	while (*str && *reg)
	{
		switch (*reg)
		{
		case '*':
			/* FIXME: '..*?..' support broken */
			while (*reg == '*' || *reg == '?') reg++;

			switch (*reg)
			{
			case '\0':
				return true;
			case '\\':
				if (*++reg == '\0') return false;
			default:
				if (case_sensitive)
				{
					while ( str = strchr(str,*reg) )
						if (str_match(str++,reg,case_sensitive)) return true;
				}
				else
				{
					char *save = (char *) str;

					while ( str = strchr(str,tolower(*reg)) )
						if (str_match(str++,reg,case_sensitive)) return true;
					str = save;
					while ( str = strchr(str,toupper(*reg)) )
						if (str_match(str++,reg,case_sensitive)) return true;
				}
				return false;
			}
			break;
		case '?':
			str++, reg++;
			break;
		case '\\':
			if (*++reg == '\0') return false;
		default:
			if (case_sensitive)
				{ if (*str++ != *reg++) return false; }
			else
				{ if ( tolower(*str++) != tolower(*reg++) ) return false; }
			break;
		}
	}

	if (*str || *reg) return false;

	return true;
}
